import math
from typing import Optional, Tuple, ClassVar

from AnyQt.QtCore import QModelIndex, QSize, Qt
from AnyQt.QtWidgets import QStyle, QStyleOptionViewItem, QApplication

from orangewidget.utils.itemdelegates import DataDelegate
from orangewidget.utils.cache import LRUCache
from orangewidget.gui import OrangeUserRole


class FixedFormatNumericColumnDelegate(DataDelegate):
    """
    A numeric delegate displaying in a fixed format.

    Parameters
    ----------
    ndecimals: int
        The number of decimals in the display
    ndigits: int
        The max number of digits in the integer part. If the model returns
        `ColumnDataSpanRole` data for a column then the `ndigits` is derived
        from that.

        .. note:: This is only used for size hinting.

    """
    ColumnDataSpanRole = next(OrangeUserRole)

    def __init__(self, *args, ndecimals=3, ndigits=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.ndecimals = ndecimals
        self.ndigits = ndigits
        self.__sh_cache = LRUCache(maxlen=200)
        self.__style = None

    def displayText(self, value, locale) -> str:
        if isinstance(value, self.RealTypes):
            return locale.toString(float(value), 'f', self.ndecimals)
        return super().displayText(value, locale)  # pragma: no cover

    def spanData(self, index: QModelIndex) -> Optional[Tuple[float, float]]:
        """
        Return the min, max numeric data values in the column that `index`
        is in.
        """
        span = self.cachedData(index, self.ColumnDataSpanRole)
        try:
            min_, max_ = span
        except (ValueError, TypeError):  # pragma: no cover
            return None
        if isinstance(min_, self.NumberTypes) and isinstance(max_, self.NumberTypes):
            return float(min_), float(max_)
        else:
            return None  # pragma: no cover

    @staticmethod
    def template(value: float, ndecimals=3) -> str:
        sign = math.copysign(1., value)
        # number of digits (integer part)
        ndigits = int(math.ceil(math.log10(abs(value) + 1)))
        template = "X" * ndigits + "." + "X" * ndecimals
        if sign == -1.:
            template = "-" + template
        return template

    def sizeHint(
            self, option: QStyleOptionViewItem, index: QModelIndex
    ) -> QSize:
        widget = option.widget
        template = self.template(-10 ** self.ndigits, self.ndecimals)
        span = self.spanData(index)
        if span is not None:
            vmin, vmax = span
            t1 = self.template(vmin, self.ndecimals)
            t2 = self.template(vmax, self.ndecimals)
            template = max((t1, t2), key=len)
        style = widget.style() if widget is not None else QApplication.style()
        # Keep ref to style wrapper. This is ugly, wrong but the wrapping of
        # C++ QStyle instance takes ~5% unless the wrapper already exists.
        self.__style = style
        opt = QStyleOptionViewItem(option)
        opt.features |= QStyleOptionViewItem.HasDisplay
        sh = QSize()
        key = option.font.key(), template
        if key not in self.__sh_cache:
            for d in map(str, range(10)):
                opt.text = template.replace("X", d)
                sh_ = style.sizeFromContents(
                    QStyle.CT_ItemViewItem, opt, QSize(), widget)
                sh = sh.expandedTo(sh_)
            self.__sh_cache[key] = sh
        else:
            sh = self.__sh_cache[key]
        return QSize(sh)


class TableDataDelegate(DataDelegate):
    """
    A DataDelegate initialized to be used with
    :class:`Orange.widgets.utils.itemmodels.TableModel`
    """
    #: Roles supplied by TableModel we want DataDelegate to use.
    DefaultRoles: ClassVar[Tuple[int, ...]] = (
        Qt.DisplayRole, Qt.TextAlignmentRole, Qt.BackgroundRole,
        Qt.ForegroundRole
    )



from unittest import TestCase

from pkg_resources import parse_version


class TestPkgResources(TestCase):
    def test_parse_version(self):
        self.assertGreater(parse_version('3.4.1'), parse_version('3.4.0'))
        self.assertGreater(parse_version('3.4.1'), parse_version('3.4.dev'))
        self.assertGreater(parse_version('3.4.1'), parse_version('3.4.1.dev'))
        self.assertLess(parse_version('3.4.1'), parse_version('3.4.2.dev'))



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring, protected-access,unsubscriptable-object
import json
import os
import tempfile
import unittest
from unittest.mock import patch, Mock

import numpy as np
from AnyQt.QtCore import Qt, QSize, QRect
from AnyQt.QtGui import QBrush

from orangewidget.tests.base import GuiTest

from Orange.data import Table, ContinuousVariable, DiscreteVariable, Domain
from Orange.preprocess.transformation import Identity
from Orange.util import color_to_hex
from Orange.widgets.utils import colorpalettes
from Orange.widgets.data import owcolor
from Orange.widgets.data.owcolor import ColorRole
from Orange.widgets.tests.base import WidgetTest


class AttrDescTest(unittest.TestCase):
    def test_name(self):
        x = ContinuousVariable("x")
        desc = owcolor.AttrDesc(x)
        self.assertEqual(desc.name, "x")
        desc.name = "y"
        self.assertEqual(desc.name, "y")
        desc.name = None
        self.assertEqual(desc.name, "x")

    def test_no_compute_value(self):
        x = ContinuousVariable("x", compute_value=lambda x: 42)
        desc = owcolor.AttrDesc(x)
        self.assertIsNone(desc.var.compute_value)

    def test_reset(self):
        x = ContinuousVariable("x")
        desc = owcolor.AttrDesc(x)
        desc.reset()
        self.assertEqual(desc.name, "x")
        desc.name = "y"
        desc.reset()
        self.assertEqual(desc.name, "x")

    def test_to_dict(self):
        x = ContinuousVariable("x")
        desc = owcolor.AttrDesc(x)
        self.assertEqual(desc.to_dict(), {})
        desc2, warns = owcolor.AttrDesc.from_dict(x, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertIsNone(desc2.new_name)

        desc.name = "y"
        self.assertEqual(desc.to_dict(), {"rename": "y"})

        desc2, warns = owcolor.AttrDesc.from_dict(x, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertEqual(desc2.new_name, "y")

        self.assertRaises(owcolor.InvalidFileFormat,
                          owcolor.AttrDesc.from_dict, x, {"rename": 42})
        self.assertRaises(owcolor.InvalidFileFormat,
                          owcolor.AttrDesc.from_dict, x, [])

        # Additional keys shouldn't cause exceptions
        owcolor.AttrDesc.from_dict(x, {"foo": 42})


class DiscAttrTest(unittest.TestCase):
    def setUp(self):
        self.var = DiscreteVariable("x", ["a", "b", "c"])
        self.desc = owcolor.DiscAttrDesc(self.var)

    def test_colors(self):
        desc = self.desc
        colors = desc.colors.copy()
        desc.set_color(2, (0, 0, 0))
        colors[2] = 0
        np.testing.assert_equal(desc.colors, colors)

    def test_values(self):
        desc = self.desc
        self.assertEqual(desc.values, ("a", "b", "c"))
        desc.set_value(1, "d")
        self.assertEqual(desc.values, ("a", "d", "c"))

    def test_create_variable(self):
        desc = self.desc
        desc.set_color(0, [1, 2, 3])
        desc.set_color(1, [4, 5, 6])
        desc.set_color(2, [7, 8, 9])
        desc.name = "z"
        desc.set_value(1, "d")
        var = desc.create_variable(self.var)
        self.assertIsInstance(var, DiscreteVariable)
        self.assertEqual(var.name, "z")
        self.assertEqual(var.values, ("a", "d", "c"))
        np.testing.assert_equal(var.colors, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertIsInstance(var.compute_value, Identity)
        self.assertIs(var.compute_value.variable, self.var)

        palette = desc.var.attributes["palette"] = object()
        var = desc.create_variable(self.var)
        self.assertIs(desc.var.attributes["palette"], palette)
        self.assertFalse(hasattr(var.attributes, "palette"))
        self.assertIsInstance(var.compute_value, Identity)
        self.assertIs(var.compute_value.variable, self.var)

    def test_reset(self):
        desc = self.desc
        desc.set_color(0, [1, 2, 3])
        desc.set_color(1, [4, 5, 6])
        desc.set_color(2, [7, 8, 9])
        desc.set_value(1, "d")
        desc.reset()
        np.testing.assert_equal(desc.colors, self.var.colors)
        self.assertEqual(desc.values, self.var.values)

    def test_to_dict(self):
        desc = owcolor.DiscAttrDesc(self.var)
        self.assertEqual(desc.to_dict(), {})
        desc2, warns = owcolor.DiscAttrDesc.from_dict(self.var, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertIsNone(desc2.new_name)
        self.assertIsNone(desc2.new_values)
        self.assertIsNone(desc2.new_colors)

        desc.name = "y"
        self.assertEqual(desc.to_dict(), {"rename": "y"})

        desc2, warns = owcolor.DiscAttrDesc.from_dict(self.var, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertEqual(desc2.new_name, "y")
        self.assertIsNone(desc2.new_values)
        self.assertIsNone(desc2.new_colors)

        desc.set_value(1, "b2")
        desc.set_color(1, [1, 2, 3])
        desc.set_color(2, [2, 3, 4])
        self.assertEqual(desc.to_dict(), {
            "rename": "y",
            "renamed_values": {"b": "b2"},
            "colors": {"a": color_to_hex(desc.colors[0]),
                       "b": "#010203",
                       "c": "#020304"}
        })

        desc2, warns = owcolor.DiscAttrDesc.from_dict(self.var, desc.to_dict())
        self.assertEqual(warns, [])
        cols = list(desc.colors)
        cols[1:] = [[1, 2, 3], [2, 3, 4]]
        np.testing.assert_equal(desc2.colors, cols)
        self.assertEqual(desc2.values, ("a", "b2", "c"))

        desc2, warns = owcolor.DiscAttrDesc.from_dict(
            self.var,
            {"rename": "y",
             "renamed_values": {"b": "b2", "d": "x"},  # d is redundant
             "colors": {"b": "#010203",
                        "c": "#020304",
                        "d": "#123456"}  # d is redundant and must be ignored
            })
        self.assertEqual(warns, [])
        cols = list(desc.colors)
        cols[1:] = [[1, 2, 3], [2, 3, 4]]
        np.testing.assert_equal(desc2.colors, cols)
        self.assertEqual(desc2.values, ("a", "b2", "c"))

    def test_from_dict_coliding_values(self):
        desc2, warns = owcolor.DiscAttrDesc.from_dict(
            self.var, {"renamed_values": {"a": "b"}})
        self.assertEqual(len(warns), 1)
        self.assertTrue("duplicate names" in warns[0])
        self.assertIsNone(desc2.new_values)
        self.assertEqual(desc2.values, ("a", "b", "c"))

        desc2, warns = owcolor.DiscAttrDesc.from_dict(
            self.var, {"renamed_values": {"a": "e", "b": "e"}})
        self.assertEqual(len(warns), 1)
        self.assertTrue("duplicate names" in warns[0])
        self.assertIsNone(desc2.new_values)
        self.assertEqual(desc2.values, ("a", "b", "c"))

        desc2, warns = owcolor.DiscAttrDesc.from_dict(
            self.var, {"renamed_values": {"a": "b", "b": "a"}})
        self.assertEqual(warns, [])
        self.assertEqual(desc2.values, ("b", "a", "c"))

    def test_from_dict_exceptions(self):
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.AttrDesc.from_dict, self.var, [])
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"rename": 42}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"colors": []}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"colors": {"a": 42}}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"colors": {4: "#000000"}}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"colors": {"a": "#00"}}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"colors": {"a": "#qwerty"}}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"renamed_values": []}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"renamed_values": {"a": 42}}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.DiscAttrDesc.from_dict, self.var, {"renamed_values": {4: "#000000"}}
        )


class ContAttrDescTest(unittest.TestCase):
    def setUp(self):
        self.var = ContinuousVariable("x")
        self.desc = owcolor.ContAttrDesc(self.var)

    def test_palette(self):
        desc = self.desc
        palette = desc.palette_name
        self.assertIsInstance(palette, str)
        desc.palette_name = "foo"
        self.assertEqual(desc.palette_name, "foo")
        desc.palette_name = None
        self.assertEqual(desc.palette_name, palette)

    def test_create_variable(self):
        desc = self.desc
        desc.name = "z"
        palette_name = _find_other_palette(
            colorpalettes.ContinuousPalettes[desc.palette_name]).name
        desc.palette_name = palette_name
        var = desc.create_variable(self.var)
        self.assertIsInstance(var, ContinuousVariable)
        self.assertEqual(var.name, "z")
        self.assertEqual(var.palette.name, palette_name)
        self.assertIsInstance(var.compute_value, Identity)
        self.assertIs(var.compute_value.variable, self.var)

        colors = desc.var.attributes["colors"] = object()
        var = desc.create_variable(self.var)
        self.assertIs(desc.var.attributes["colors"], colors)
        self.assertFalse(hasattr(var.attributes, "colors"))
        self.assertIsInstance(var.compute_value, Identity)
        self.assertIs(var.compute_value.variable, self.var)

    def test_reset(self):
        desc = self.desc
        palette_name = desc.palette_name
        desc.palette_name = _find_other_palette(
            colorpalettes.ContinuousPalettes[palette_name]).name
        desc.reset()
        np.testing.assert_equal(desc.palette_name, palette_name)

    def test_to_dict(self):
        x = ContinuousVariable("x")
        desc = owcolor.ContAttrDesc(x)
        self.assertEqual(desc.to_dict(), {})
        desc2, warns = owcolor.ContAttrDesc.from_dict(x, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertIsNone(desc2.new_name)
        self.assertIsNone(desc2.new_palette_name)

        desc.name = "y"
        self.assertEqual(desc.to_dict(), {"rename": "y"})

        desc2, warns = owcolor.ContAttrDesc.from_dict(x, desc.to_dict())
        self.assertEqual(warns, [])
        self.assertEqual(desc2.new_name, "y")
        self.assertIsNone(desc2.new_palette_name)

        desc = owcolor.ContAttrDesc(x)
        desc.palette_name = "linear_viridis"
        self.assertEqual(desc.to_dict(), {"colors": "linear_viridis"})

    def test_from_dict_exceptions(self):
        x = ContinuousVariable("x")
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.ContAttrDesc.from_dict, x, []
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.ContAttrDesc.from_dict, x, {"colors": 42}
        )
        self.assertRaises(
            owcolor.InvalidFileFormat,
            owcolor.ContAttrDesc.from_dict, x, {"colors": "no such palette"}
        )


class BaseTestColorTableModel:
    def test_row_count(self):
        model = self.model

        self.assertEqual(model.rowCount(), 0)
        model.set_data(self.descs)
        self.assertEqual(model.rowCount(), len(self.descs))
        self.assertEqual(model.rowCount(self.model.index(0, 0)), 0)

    def test_data(self):
        self.model.set_data(self.descs)
        data = self.model.data

        index = self.model.index(1, 0)
        self.assertEqual(data(index, Qt.DisplayRole), self.descs[1].name)
        self.assertEqual(data(index, Qt.EditRole), self.descs[1].name)
        self.assertTrue(data(index, Qt.FontRole).bold())
        self.assertTrue(data(index, Qt.TextAlignmentRole) & Qt.AlignRight)

        self.descs[1].name = "bar"
        self.assertEqual(data(index), "bar")

        index = self.model.index(2, 0)
        self.assertEqual(data(index, Qt.DisplayRole), self.descs[2].name)

    def test_set_data(self):
        emit = Mock()
        try:
            self.model.dataChanged.connect(emit)
            self.model.set_data(self.descs)
            data = self.model.data
            setData = self.model.setData

            index = self.model.index(1, 0)
            assert self.descs[1].name != "foo"
            self.assertFalse(setData(index, "foo", Qt.DisplayRole))
            emit.assert_not_called()
            self.assertEqual(data(index, Qt.DisplayRole), self.descs[1].name)
            self.assertTrue(setData(index, "foo", Qt.EditRole))
            emit.assert_called()
            self.assertEqual(data(index, Qt.DisplayRole), "foo")
            self.assertEqual(self.descs[1].name, "foo")
        finally:
            self.model.dataChanged.disconnect(emit)

    def test_reset(self):
        for desc in self.descs:
            desc.reset = Mock()
        self.model.set_data(self.descs)
        self.model.reset()
        for desc in self.descs:
            desc.reset.assert_called()


class TestDiscColorTableModel(GuiTest, BaseTestColorTableModel):
    def setUp(self):
        x = DiscreteVariable("x", list("abc"))
        y = DiscreteVariable("y", list("def"))
        z = DiscreteVariable("z", list("ghijk"))
        self.descs = [owcolor.DiscAttrDesc(v) for v in (x, y, z)]
        self.model = owcolor.DiscColorTableModel()

    def test_column_count(self):
        model = self.model

        self.assertEqual(model.columnCount(), 1)
        model.set_data(self.descs[:2])
        self.assertEqual(model.columnCount(), 4)
        model.set_data(self.descs)
        self.assertEqual(model.columnCount(), 6)

        self.assertEqual(model.columnCount(model.index(0, 0)), 0)

    def test_data(self):
        super().test_data()

        model = self.model

        self.assertIsNone(model.data(model.index(0, 4)))

        index = model.index(1, 2)
        self.assertEqual(model.data(index, Qt.DisplayRole), "e")
        self.assertEqual(model.data(index, Qt.EditRole), "e")
        font = model.data(index, Qt.FontRole)
        self.assertTrue(font is None or not font.bold())

        var_colors = self.descs[1].var.colors[1]
        color = model.data(index, Qt.DecorationRole)
        np.testing.assert_equal(color.getRgb()[:3], var_colors)

        color = model.data(index, owcolor.ColorRole)
        np.testing.assert_equal(color, var_colors)

        self.assertEqual(
            model.data(index, Qt.ToolTipRole), color_to_hex(var_colors))

        self.assertIsNone(model.data(model.index(0, 4)))

        index = model.index(2, 5)
        self.assertEqual(model.data(index, Qt.DisplayRole), "k")
        self.assertEqual(model.data(index, Qt.EditRole), "k")
        font = model.data(index, Qt.FontRole)
        self.assertTrue(font is None or not font.bold())

        var_colors = self.descs[2].var.colors[4]
        color = model.data(index, Qt.DecorationRole)
        np.testing.assert_equal(color.getRgb()[:3], var_colors)

        color = model.data(index, owcolor.ColorRole)
        np.testing.assert_equal(color, var_colors)

        self.assertEqual(
            model.data(index, Qt.ToolTipRole), color_to_hex(var_colors))

        self.descs[2].set_value(4, "foo")
        self.assertEqual(model.data(index, Qt.DisplayRole), "foo")

    def test_set_data(self):
        super().test_set_data()

        model = self.model
        emit = Mock()
        try:
            model.dataChanged.connect(emit)

            index = model.index(2, 5)

            self.assertEqual(model.data(index, Qt.DisplayRole), "k")
            self.assertEqual(model.data(index, Qt.EditRole), "k")
            self.assertFalse(model.setData(index, "foo", Qt.DisplayRole))
            emit.assert_not_called()
            self.assertEqual(model.data(index, Qt.DisplayRole), "k")
            self.assertTrue(model.setData(index, "foo", Qt.EditRole))
            emit.assert_called()
            emit.reset_mock()
            self.assertEqual(model.data(index, Qt.DisplayRole), "foo")
            self.assertEqual(self.descs[2].values, ("g", "h", "i", "j", "foo"))

            new_color = [0, 1, 2]
            self.assertTrue(model.setData(index, new_color + [255], ColorRole))
            emit.assert_called()
            emit.reset_mock()
            color = model.data(index, Qt.DecorationRole)
            rgb = [color.red(), color.green(), color.blue()]
            self.assertEqual(rgb, new_color)

            color = model.data(index, owcolor.ColorRole)
            self.assertEqual(list(color), new_color)

            self.assertEqual(
                model.data(index, Qt.ToolTipRole), color_to_hex(new_color))

            np.testing.assert_equal(self.descs[2].colors[4], rgb)
        finally:
            model.dataChanged.disconnect(emit)


def _find_other_palette(initial):
    for palette in colorpalettes.ContinuousPalettes.values():
        if palette.name != initial.name:
            return palette
    return None  # pragma: no cover


class TestContColorTableModel(GuiTest, BaseTestColorTableModel):
    def setUp(self):
        z = ContinuousVariable("z")
        w = ContinuousVariable("w")
        u = ContinuousVariable("u")
        self.descs = [owcolor.ContAttrDesc(v) for v in (z, w, u)]
        self.model = owcolor.ContColorTableModel()

    def test_column_count(self):
        model = self.model

        model.set_data(self.descs)
        self.assertEqual(model.columnCount(), 3)
        self.assertEqual(model.columnCount(model.index(0, 0)), 0)

    def test_data(self):
        super().test_data()

        model = self.model
        index = model.index(1, 1)
        palette = colorpalettes.ContinuousPalettes[self.descs[1].palette_name]
        self.assertEqual(model.data(index, Qt.ToolTipRole),
                         palette.friendly_name)
        self.assertEqual(model.data(index, ColorRole), palette)
        with patch.object(palette, "color_strip") as color_strip:
            strip = model.data(index, owcolor.StripRole)
            self.assertIs(strip, color_strip.return_value)
            color_strip.assert_called_with(128, 16)
        self.assertIsInstance(model.data(index, Qt.SizeHintRole), QSize)
        self.assertIsNone(model.data(index, Qt.FontRole))

        palette = _find_other_palette(self.descs[1])
        self.descs[1].palette_name = palette.name
        self.assertIs(model.data(index, ColorRole), palette)

        index = self.model.index(1, 2)
        self.assertIsNone(model.data(index, Qt.ToolTipRole))
        self.assertIsInstance(model.data(index, Qt.SizeHintRole), QSize)
        self.assertIsInstance(model.data(index, Qt.ForegroundRole), QBrush)
        self.assertIsNone(model.data(index, Qt.DisplayRole))
        model.set_mouse_row(0)
        self.assertIsNone(model.data(index, Qt.DisplayRole))
        model.set_mouse_row(1)
        self.assertEqual(model.data(index, Qt.DisplayRole), "Copy to all")

    def test_set_data(self):
        super().test_set_data()

        model = self.model
        index = model.index(1, 1)
        index2 = model.index(2, 1)
        initial = model.data(index, ColorRole)
        initial2 = model.data(index, ColorRole)
        assert initial.name == initial2.name
        palette = _find_other_palette(initial)

        emit = Mock()
        try:
            model.dataChanged.connect(emit)

            self.assertFalse(model.setData(index, None, Qt.DisplayRole))
            emit.assert_not_called()

            self.assertTrue(model.setData(index, palette, ColorRole))
            emit.assert_called()
            self.assertIs(model.data(index2, ColorRole), initial2)

            self.assertEqual(model.data(index, Qt.ToolTipRole),
                             palette.friendly_name)
            self.assertEqual(model.data(index, ColorRole), palette)
            self.assertEqual(self.descs[1].palette_name, palette.name)
            with patch.object(palette, "color_strip") as color_strip:
                strip = model.data(index, owcolor.StripRole)
                self.assertIs(strip, color_strip.return_value)
                color_strip.assert_called_with(128, 16)
        finally:
            model.dataChanged.disconnect(emit)

    def test_copy_to_all(self):
        super().test_set_data()

        model = self.model
        index = model.index(1, 1)
        initial = model.data(index, ColorRole)
        palette = _find_other_palette(initial)

        emit = Mock()
        try:
            model.dataChanged.connect(emit)
            model.setData(index, palette, ColorRole)
            emit.assert_called()
            emit.reset_mock()

            model.copy_to_all(index)
            emit.assert_called_once()
            for row, desc in enumerate(self.descs):
                self.assertEqual(
                    model.data(model.index(row, 1), ColorRole).name,
                    palette.name)
                self.assertEqual(desc.palette_name, palette.name)
        finally:
            model.dataChanged.disconnect(emit)


class TestColorStripDelegate(GuiTest):
    def setUp(self):
        z = ContinuousVariable("z")
        w = ContinuousVariable("w")
        u = ContinuousVariable("u")
        self.descs = [owcolor.ContAttrDesc(v) for v in (z, w, u)]
        self.model = owcolor.ContColorTableModel()
        self.model.set_data(self.descs)
        self.table = owcolor.ContinuousTable(self.model)

    def test_color_combo(self):
        model = self.model
        index = model.index(1, 1)
        initial = model.data(index, ColorRole)
        palette = _find_other_palette(initial)
        model.setData(index, palette, ColorRole)
        self.assertEqual(self.descs[1].palette_name, palette.name)

        combo = self.table.color_delegate.createEditor(None, Mock(), index)
        self.assertEqual(combo.currentText(), palette.friendly_name)
        palette = _find_other_palette(palette)
        combo.setCurrentIndex(combo.findText(palette.friendly_name))
        self.assertEqual(self.descs[1].palette_name, palette.name)

        with patch.object(self.table, "closeEditor") as closeEditor:
            combo.hidePopup()
            closeEditor.assert_called()

    @patch.object(owcolor.HorizontalGridDelegate, "paint")
    def test_paint(self, _):
        model = self.model
        index = model.index(1, 1)
        painter = Mock()
        option = Mock()
        option.rect = QRect(10, 20, 30, 40)
        index.data = Mock()
        index.data.return_value = Mock()
        index.data.return_value.height = Mock(return_value=42)
        self.table.color_delegate.paint(painter, option, index)
        self.assertIs(painter.drawPixmap.call_args[0][2],
                      index.data.return_value)


class TestOWColor(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(owcolor.OWColor)
        self.iris = Table("iris")

    def test_reuse_old_settings(self):
        self.send_signal(self.widget.Inputs.data, self.iris)

        assert isinstance(self.widget, owcolor.OWColor)
        self.widget.saveSettings()

        w = self.create_widget(owcolor.OWColor, reset_default_settings=False)
        self.send_signal(self.widget.Inputs.data, self.iris, widget=w)

    def test_invalid_input_colors(self):
        # Handling of invalid value in deprecated 'colors': warn, but don't fail
        with self.assertWarns(DeprecationWarning):
            a = ContinuousVariable("a")
            a.attributes["colors"] = "invalid"
            t = Table.from_domain(Domain([a]))
            self.send_signal(self.widget.Inputs.data, t)

    def test_unconditional_commit_on_new_signal(self):
        with patch.object(self.widget.commit, 'now') as commit:
            self.widget.auto_apply = False
            commit.reset_mock()
            self.send_signal(self.widget.Inputs.data, self.iris)
            commit.assert_called()

    def test_commit_on_data_changed(self):
        widget = self.widget
        model = widget.cont_model
        self.send_signal(widget.Inputs.data, self.iris)
        with patch.object(widget.commit, 'deferred') as commit:
            commit.reset_mock()
            model.setData(model.index(0, 0), "y", Qt.EditRole)
            commit.assert_called()

    def test_lose_data(self):
        widget = self.widget
        send = widget.Outputs.data.send = Mock()

        self.send_signal(widget.Inputs.data, self.iris)
        send.assert_called_once()
        self.assertEqual(widget.disc_model.rowCount(), 1)
        self.assertEqual(widget.cont_model.rowCount(), 4)

        send.reset_mock()
        self.send_signal(widget.Inputs.data, None)
        send.assert_called_with(None)
        self.assertEqual(widget.disc_model.rowCount(), 0)
        self.assertEqual(widget.cont_model.rowCount(), 0)

    def test_model_content(self):
        widget = self.widget
        data = Table("heart_disease")
        self.send_signal(widget.Inputs.data, data)

        dm = widget.disc_model
        self.assertEqual(
            [dm.data(dm.index(i, 0)) for i in range(dm.rowCount())],
            [var.name for var in data.domain.variables if var.is_discrete]
        )

        cm = widget.disc_model
        self.assertEqual(
            [dm.data(cm.index(i, 0)) for i in range(cm.rowCount())],
            [var.name for var in data.domain.variables if var.is_discrete]
        )

    def test_report(self):
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.data, Table("zoo"))
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.data, None)
        self.widget.send_report()

    def test_string_variables(self):
        self.send_signal(self.widget.Inputs.data, Table("zoo"))

    def test_changed_compute_value(self):
        # test a bug where the widget did not register changes in compute_value
        # because it reused an old context
        w = self.widget
        domain1 = Domain([ContinuousVariable("x", compute_value=lambda _: 1)])
        data1 = self.iris.transform(domain1)
        self.send_signal(w.Inputs.data, data1)
        outp = self.get_output(w.Outputs.data)
        np.testing.assert_array_equal(outp, 1)
        domain2 = Domain([ContinuousVariable("x", compute_value=lambda _: 2)])
        data2 = self.iris.transform(domain2)
        self.send_signal(w.Inputs.data, data2)
        outp = self.get_output(w.Outputs.data)
        np.testing.assert_array_equal(outp, 2)

    def test_reset(self):
        self.send_signal(self.widget.Inputs.data, self.iris)
        cont_model = self.widget.cont_model
        disc_model = self.widget.disc_model
        cont_model.setData(cont_model.index(0, 0), "a", Qt.EditRole)
        disc_model.setData(disc_model.index(0, 0), "b", Qt.EditRole)
        outp = self.get_output(self.widget.Outputs.data)
        self.assertEqual(outp.domain[0].name, "a")
        self.assertEqual(outp.domain.class_var.name, "b")
        self.widget.reset()
        outp = self.get_output(self.widget.Outputs.data)
        self.assertEqual(
            outp.domain[0].name, self.iris.domain[0].name)
        self.assertEqual(
            outp.domain.class_var.name, self.iris.domain.class_var.name)

    def test_save(self):
        self.widget._save_var_defs = Mock()
        with patch.object(owcolor.QFileDialog, "getSaveFileName",
                          return_value=("", "")):
            self.widget.save()
            self.widget._save_var_defs.assert_not_called()
        with patch.object(owcolor.QFileDialog, "getSaveFileName",
                          return_value=("foo", "bar")):
            self.widget.save()
            self.widget._save_var_defs.assert_called_with("foo")

    def test_save_low(self):
        descA = owcolor.DiscAttrDesc(
            DiscreteVariable("varA", values=tuple("abc")))
        descA.name = "a2"
        descB = owcolor.DiscAttrDesc(
            DiscreteVariable("varB", values=tuple("abc")))
        descB.set_value(1, "X")
        descC = owcolor.ContAttrDesc(ContinuousVariable("varC"))
        descC.name = "c2"
        descD = owcolor.ContAttrDesc(ContinuousVariable("varD"))
        descD.palette_name = "linear_viridis"
        descE = owcolor.ContAttrDesc(ContinuousVariable("varE"))
        self.widget.disc_descs = [descA, descB]
        self.widget.cont_descs = [descC, descD, descE]

        with tempfile.TemporaryDirectory() as path:
            fname = os.path.join(path, "foo.colors")
            self.widget._save_var_defs(fname)
            with open(fname) as f:
                js = json.load(f)
            self.assertEqual(js["categorical"],
                             {"varA": {"rename": "a2"},
                              "varB": {"renamed_values": {"b": "X"}}})
            self.assertEqual(js["numeric"],
                             {"varC": {"rename": "c2"},
                              "varD": {"colors": "linear_viridis"}})

    @patch("Orange.widgets.data.owcolor.QMessageBox.critical")
    def test_load(self, msg_box):
        self.widget._parse_var_defs = Mock()
        self.widget._save_var_defs = Mock()
        with patch.object(owcolor.QFileDialog, "getOpenFileName",
                          return_value=("", "")):
            with patch("builtins.open"):
                self.widget.load()
                open.assert_not_called()
                self.widget._parse_var_defs.assert_not_called()

        with patch.object(owcolor.QFileDialog, "getOpenFileName",
                          return_value=("foo.colors", "*.colors")):
            with patch("json.load") as json_load, \
                    patch("builtins.open", side_effect=IOError):
                self.widget.load()
                msg_box.assert_called()
                msg_box.reset_mock()
                json_load.assert_not_called()
                self.widget._parse_var_defs.assert_not_called()

        with patch.object(owcolor.QFileDialog, "getOpenFileName",
                          return_value=("foo.colors", "*.colors")):
            with patch("json.load", side_effect=json.JSONDecodeError("err", "d", 42)), \
                    patch("builtins.open"):
                self.widget.load()
                msg_box.assert_called()
                msg_box.reset_mock()
                self.widget._parse_var_defs.assert_not_called()

        with patch.object(owcolor.QFileDialog, "getOpenFileName",
                          return_value=("foo.colors", "*.colors")):
            with patch("json.load"), patch("builtins.open"):
                self.widget.load()
                msg_box.assert_not_called()
                msg_box.reset_mock()
                self.widget._parse_var_defs.assert_called_with(json.load.return_value)

    @patch("Orange.widgets.data.owcolor.QMessageBox.warning")
    def test_load_ignore_warning(self, msg_box):
        self.widget._parse_var_defs(dict(categorical={}, numeric={}))
        msg_box.assert_not_called()

        no_change = dict(renamed_values={}, colors={})
        for names, message in (
                (("foo",),
                 "'foo'"),
                (("foo", "bar"),
                 "'foo' and 'bar'"),
                (("foo", "bar", "baz"),
                 "'foo', 'bar' and 'baz'"),
                (("foo", "bar", "baz", "qux"),
                 "'foo', 'bar', 'baz' and 'qux'"),
                (("foo", "bar", "baz", "qux", "quux"),
                 "'foo', 'bar', 'baz', 'qux' and 'quux'"),
                (("foo", "bar", "baz", "qux", "quux", "corge"),
                 "'foo', 'bar', 'baz', 'qux' and 2 other"),
                (("foo", "bar", "baz", "qux", "quux", "corge", "grault"),
                 "'foo', 'bar', 'baz', 'qux' and 3 other")):
            self.widget._parse_var_defs(dict(
                categorical=dict.fromkeys(names, no_change),
                numeric={}))
            self.assertIn(message, msg_box.call_args[0][2])

    def _create_descs(self):
        disc_vars = [DiscreteVariable(f"var{c}", values=("a", "b", "c"))
                     for c in "AB"]
        cont_vars = [ContinuousVariable(f"var{c}") for c in "CDE"]
        self.widget.disc_descs = [owcolor.DiscAttrDesc(v) for v in disc_vars]
        self.widget.cont_descs = [owcolor.ContAttrDesc(v) for v in cont_vars]
        return disc_vars, cont_vars

    def test_parse_var_defs(self):
        js = {"categorical": {"varA": {"rename": "a2"},
                              "varB": {"renamed_values": {"b": "X"}}},
              "numeric": {"varC": {"rename": "c2"},
                          "varD": {"colors": "linear_viridis"}}}

        self._create_descs()
        descE = self.widget.cont_descs[-1]

        self.widget._parse_var_defs(js)
        self.assertEqual(len(self.widget.disc_descs), 2)
        descA = self.widget.disc_descs[0]
        self.assertEqual(descA.name, "a2")
        self.assertIsNone(descA.new_values)
        self.assertIsNone(descA.new_colors)

        descB = self.widget.disc_descs[1]
        self.assertIsNone(descB.new_name)
        self.assertEqual(descB.new_values, ["a", "X", "c"])
        self.assertIsNone(descB.new_colors)

        self.assertEqual(len(self.widget.cont_descs), 3)
        descC = self.widget.cont_descs[0]
        self.assertEqual(descC.name, "c2")
        self.assertIsNone(descC.new_palette_name)

        descD = self.widget.cont_descs[1]
        self.assertIsNone(descD.new_name)
        self.assertEqual(descD.new_palette_name, "linear_viridis")

        self.assertIs(self.widget.cont_descs[2], descE)

    def test_parse_var_defs_invalid(self):
        self.assertRaises(
            owcolor.InvalidFileFormat,
            self.widget._parse_var_defs, 42)
        self.assertRaises(
            owcolor.InvalidFileFormat,
            self.widget._parse_var_defs,
            {"categorical": {"a": 42}, "numeric": {}})
        self.assertRaises(
            owcolor.InvalidFileFormat,
            self.widget._parse_var_defs,
            {"categorical": {"a": {"rename": 4}}, "numeric": {}})
        self.assertRaises(
            owcolor.InvalidFileFormat,
            self.widget._parse_var_defs,
            {"categorical": {42: {"rename": "b"}}, "numeric": {}})

    @patch("Orange.widgets.data.owcolor.QMessageBox.warning")
    def test_parse_var_defs_shows_warnings(self, msg_box):
        self._create_descs()
        self.widget._parse_var_defs(
            {"categorical": {"varA": {"renamed_values": {"a": "b"}}},
             "numeric": {}})
        msg_box.assert_called()
        self.assertTrue("duplicate names" in msg_box.call_args[0][2])

    @patch("Orange.widgets.data.owcolor.QMessageBox.warning")
    def test_parse_var_defs_no_rename(self, msg_box):
        self._create_descs()

        self.widget._parse_var_defs(
            {"categorical": {"varA": {"rename": "varB"}},
             "numeric": {}})
        msg_box.assert_called()
        self.assertTrue("duplicated names" in msg_box.call_args[0][2])
        msg_box.reset_mock()

        self.widget._parse_var_defs(
            {"categorical": {"varA": {"rename": "X"}},
             "numeric": {"varD": {"rename": "X"}}})
        msg_box.assert_called()
        self.assertTrue("duplicated names" in msg_box.call_args[0][2])
        msg_box.reset_mock()

        self.widget._parse_var_defs(
            {"categorical": {"varA": {"rename": "varD"}},
             "numeric": {"varD": {"rename": "varA"}}})
        msg_box.assert_not_called()


if __name__ == "__main__":
    unittest.main()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring
import unittest
from unittest.mock import patch

from scipy.sparse import csr_matrix

from AnyQt.QtCore import Qt, QModelIndex
from AnyQt.QtTest import QSignalSpy
from AnyQt.QtGui import QBrush, QColor

from orangewidget.tests.base import GuiTest

from Orange.data import \
    Domain, Table, Value, \
    ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable
from Orange.statistics.basic_stats import BasicStats
from Orange.widgets.utils import colorpalettes
from Orange.widgets.utils.itemmodels import \
    PyTableModel, PyListModel, PyListModelTooltip,\
    VariableListModel, DomainModel, ContinuousPalettesModel, \
    TableModel, _as_contiguous_range
from Orange.widgets.gui import TableVariable
from Orange.tests import test_filename
from Orange.tests.sql.base import DataBaseTest as dbt
from Orange.data.sql.table import SqlTable


class TestUtils(unittest.TestCase):
    def test_as_contiguous_range(self):
        self.assertEqual(_as_contiguous_range(slice(1, 8), 20), (1, 8, 1))
        self.assertEqual(_as_contiguous_range(slice(1, 8), 6), (1, 6, 1))
        self.assertEqual(_as_contiguous_range(slice(8, 1, -1), 6), (2, 6, 1))
        self.assertEqual(_as_contiguous_range(slice(8), 6), (0, 6, 1))
        self.assertEqual(_as_contiguous_range(slice(8, None, -1), 6), (0, 6, 1))
        self.assertEqual(_as_contiguous_range(slice(7, None, -1), 9), (0, 8, 1))
        self.assertEqual(_as_contiguous_range(slice(None, None, -1), 9),
                         (0, 9, 1))


class TestPyTableModel(unittest.TestCase):
    def setUp(self):
        self.model = PyTableModel([[1, 4],
                                   [2, 3]])

    def test_init(self):
        self.model = PyTableModel()
        self.assertEqual(self.model.rowCount(), 0)

    def test_init_wrap_empty(self):
        # pylint: disable=protected-access
        t = []
        model = PyTableModel(t)
        self.assertIs(model._table, t)
        t.append([1, 2, 3])
        self.assertEqual(list(model), [[1, 2, 3]])

    def test_rowCount(self):
        self.assertEqual(self.model.rowCount(), 2)
        self.assertEqual(len(self.model), 2)

    def test_columnCount(self):
        self.assertEqual(self.model.columnCount(), 2)

    def test_data(self):
        mi = self.model.index(0, 0)
        self.assertEqual(self.model.data(mi), '1')
        self.assertEqual(self.model.data(mi, Qt.EditRole), 1)

    def test_editable(self):
        editable_model = PyTableModel([[0]], editable=True)
        self.assertFalse(bool(self.model.flags(self.model.index(0, 0)) & Qt.ItemIsEditable))
        self.assertTrue(bool(editable_model.flags(editable_model.index(0, 0)) & Qt.ItemIsEditable))

    def test_sort(self):
        self.model.sort(1)
        self.assertEqual(self.model.index(0, 0).data(Qt.EditRole), 2)

    def test_setHeaderLabels(self):
        self.model.setHorizontalHeaderLabels(['Col 1', 'Col 2'])
        self.assertEqual(self.model.headerData(1, Qt.Horizontal), 'Col 2')
        self.assertEqual(self.model.headerData(1, Qt.Vertical), 2)

    def test_removeRows(self):
        self.model.removeRows(0, 1)
        self.assertEqual(len(self.model), 1)
        self.assertEqual(self.model[0][1], 3)

    def test_removeColumns(self):
        self.model.removeColumns(0, 1)
        self.assertEqual(self.model.columnCount(), 1)
        self.assertEqual(self.model[1][0], 3)

    def test_insertRows(self):
        self.model.insertRows(0, 1)
        self.assertEqual(self.model[1][0], 1)

    def test_insertColumns(self):
        self.model.insertColumns(0, 1)
        self.assertEqual(self.model[0], ['', 1, 4])

    def test_wrap(self):
        self.model.wrap([[0]])
        self.assertEqual(self.model.rowCount(), 1)
        self.assertEqual(self.model.columnCount(), 1)

    def test_clear(self):
        self.model.clear()
        self.assertEqual(self.model.rowCount(), 0)

    def test_append(self):
        self.model.append([5, 6])
        self.assertEqual(self.model[2][1], 6)
        self.assertEqual(self.model.rowCount(), 3)

    def test_extend(self):
        self.model.extend([[5, 6]])
        self.assertEqual(self.model[2][1], 6)
        self.assertEqual(self.model.rowCount(), 3)

    def test_insert(self):
        self.model.insert(0, [5, 6])
        self.assertEqual(self.model[0][1], 6)
        self.assertEqual(self.model.rowCount(), 3)

    def test_remove(self):
        self.model.remove([2, 3])
        self.assertEqual(self.model.rowCount(), 1)

    def test_other_roles(self):
        self.model.append([2, 3])
        self.model.setData(self.model.index(2, 0),
                           Qt.AlignCenter,
                           Qt.TextAlignmentRole)
        del self.model[1]
        self.assertTrue(Qt.AlignCenter &
                        self.model.data(self.model.index(1, 0),
                                        Qt.TextAlignmentRole))

    def test_set_iten_slice(self):
        self.model[:1] = [[10, 11], [12, 13], [14, 15]]
        self.assertEqual(list(self.model), [[10, 11], [12, 13], [14, 15], [2, 3]])

        self.model[1:3] = []
        self.assertEqual(list(self.model), [[10, 11], [2, 3]])

        self.model[:] = [[20, 21]]
        self.assertEqual(list(self.model), [[20, 21]])

        self.model[1:] = [[10, 11], [2, 3]]
        self.assertEqual(list(self.model), [[20, 21], [10, 11], [2, 3]])

    def test_emits_column_changes_on_row_insert(self):
        inserted = []
        removed = []
        model = PyTableModel()
        model.columnsInserted.connect(inserted.append)
        model.columnsRemoved.connect(removed.append)
        inserted = QSignalSpy(model.columnsInserted)
        removed = QSignalSpy(model.columnsRemoved)
        model.append([2])
        self.assertEqual(list(inserted)[-1][1:], [0, 0])
        model.append([2, 3])
        self.assertEqual(list(inserted)[-1][1:], [1, 1])
        del model[:]
        self.assertEqual(list(removed)[-1][1:], [0, 1])
        model.extend([[0, 1], [0, 2]])
        self.assertEqual(list(inserted)[-1][1:], [0, 1])
        model.clear()
        self.assertEqual(list(removed)[-1][1:], [0, 1])
        model[:] = [[1], [2]]
        self.assertEqual(list(inserted)[-1][1:], [0, 0])


class TestVariableListModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.disc = DiscreteVariable("gender", values=("M", "F"))
        cls.cont = ContinuousVariable("age")
        cls.string = StringVariable("name")
        cls.time = TimeVariable("birth")
        cls.model = VariableListModel([
            cls.cont, None, "Foo", cls.disc, cls.string, cls.time])

    def test_placeholder(self):
        model = self.model
        self.assertEqual(model.data(model.index(1)), "None")
        model.placeholder = "Bar"
        self.assertEqual(model.data(model.index(1)), "Bar")
        model.placeholder = "None"

    def test_displayrole(self):
        data, index = self.model.data, self.model.index
        self.assertEqual(data(index(0)), "age")
        self.assertEqual(data(index(1)), "None")
        self.assertEqual(data(index(2)), "Foo")
        self.assertEqual(data(index(3)), "gender")
        self.assertEqual(data(index(4)), "name")
        self.assertEqual(data(index(5)), "birth")

    def test_tooltip(self):
        def get_tooltip(i):
            return self.model.data(self.model.index(i), Qt.ToolTipRole)

        text = get_tooltip(0)
        self.assertIn("age", text)
        self.assertIn("Numeric", text)

        self.assertIsNone(get_tooltip(1))
        self.assertIsNone(get_tooltip(2))

        text = get_tooltip(3)
        self.assertIn("gender", text)
        self.assertIn("M", text)
        self.assertIn("F", text)
        self.assertIn("2", text)
        self.assertIn("Categorical", text)

        text = get_tooltip(4)
        self.assertIn("name", text)
        self.assertIn("Text", text)

        text = get_tooltip(5)
        self.assertIn("birth", text)
        self.assertIn("Time", text)

        self.cont.attributes = {"foo": "bar"}
        text = get_tooltip(0)
        self.assertIn("foo", text)
        self.assertIn("bar", text)

    def test_table_variable(self):
        self.assertEqual(
            [self.model.data(self.model.index(i), TableVariable)
             for i in range(self.model.rowCount())],
            [self.cont, None, None, self.disc, self.string, self.time])

    def test_other_roles(self):
        with patch.object(PyListModel, "data") as data:
            index = self.model.index(0)
            _ = self.model.data(index, Qt.BackgroundRole)
            self.assertEqual(data.call_args[0][1:], (index, Qt.BackgroundRole))

    def test_invalid_index(self):
        self.assertIsNone(self.model.data(self.model.index(0).parent()))


class TestDomainModel(unittest.TestCase):
    def test_init_with_single_section(self):
        model = DomainModel(order=DomainModel.CLASSES)
        self.assertEqual(model.order, (DomainModel.CLASSES, ))

    def test_separators(self):
        attrs = [ContinuousVariable(n) for n in "abg"]
        classes = [ContinuousVariable(n) for n in "deh"]
        metas = [ContinuousVariable(n) for n in "ijf"]

        model = DomainModel()
        sep = [model.Separator]
        model.set_domain(Domain(attrs, classes, metas))
        self.assertEqual(list(model), classes + sep + metas + sep + attrs)

        model = DomainModel()
        model.set_domain(Domain(attrs, [], metas))
        self.assertEqual(list(model), metas + sep + attrs)

        model = DomainModel()
        model.set_domain(Domain([], [], metas))
        self.assertEqual(list(model), metas)

        model = DomainModel(placeholder="foo")
        model.set_domain(Domain([], [], metas))
        self.assertEqual(list(model), [None] + sep + metas)

        model = DomainModel(placeholder="foo")
        model.set_domain(Domain(attrs, [], metas))
        self.assertEqual(list(model), [None] + sep + metas + sep + attrs)

    def test_placeholder_placement(self):
        model = DomainModel(placeholder="foo")
        sep = model.Separator
        self.assertEqual(model.order, (None, sep) + model.SEPARATED)

        model = DomainModel(order=("bar", ), placeholder="foo")
        self.assertEqual(model.order, (None, "bar"))

        model = DomainModel(order=("bar", None, "baz"), placeholder="foo")
        self.assertEqual(model.order, ("bar", None, "baz"))

        model = DomainModel(order=("bar", sep, "baz"),
                            placeholder="foo")
        self.assertEqual(model.order, (None, sep, "bar", sep, "baz"))

    def test_subparts(self):
        attrs = [ContinuousVariable(n) for n in "abg"]
        classes = [ContinuousVariable(n) for n in "deh"]
        metas = [ContinuousVariable(n) for n in "ijf"]

        m = DomainModel
        sep = m.Separator
        model = DomainModel(
            order=(m.ATTRIBUTES | m.METAS, sep, m.CLASSES))
        model.set_domain(Domain(attrs, classes, metas))
        self.assertEqual(list(model), attrs + metas + [sep] + classes)

        m = DomainModel
        sep = m.Separator
        model = DomainModel(
            order=(m.ATTRIBUTES | m.METAS, sep, m.CLASSES),
            alphabetical=True)
        model.set_domain(Domain(attrs, classes, metas))
        self.assertEqual(list(model),
                         sorted(attrs + metas, key=lambda x: x.name) +
                         [sep] +
                         sorted(classes, key=lambda x: x.name))

    def test_filtering(self):
        cont = [ContinuousVariable(n) for n in "abc"]
        disc = [DiscreteVariable(n) for n in "def"]
        attrs = cont + disc

        model = DomainModel(valid_types=(ContinuousVariable, ))
        model.set_domain(Domain(attrs))
        self.assertEqual(list(model), cont)

        model = DomainModel(valid_types=(DiscreteVariable, ))
        model.set_domain(Domain(attrs))
        self.assertEqual(list(model), disc)

        disc[0].attributes["hidden"] = True
        model.set_domain(Domain(attrs))
        self.assertEqual(list(model), disc[1:])

        model = DomainModel(valid_types=(DiscreteVariable, ),
                            skip_hidden_vars=False)
        model.set_domain(Domain(attrs))
        self.assertEqual(list(model), disc)

    def test_no_separators(self):
        """
        GH-2697
        """
        attrs = [ContinuousVariable(n) for n in "abg"]
        classes = [ContinuousVariable(n) for n in "deh"]
        metas = [ContinuousVariable(n) for n in "ijf"]

        model = DomainModel(order=DomainModel.SEPARATED, separators=False)
        model.set_domain(Domain(attrs, classes, metas))
        self.assertEqual(list(model), classes + metas + attrs)

        model = DomainModel(order=DomainModel.SEPARATED, separators=True)
        model.set_domain(Domain(attrs, classes, metas))
        self.assertEqual(
            list(model),
            classes + [PyListModel.Separator] + metas + [PyListModel.Separator] + attrs)

    def test_read_only(self):
        model = DomainModel()
        domain = Domain([ContinuousVariable(x) for x in "abc"])
        model.set_domain(domain)
        index = model.index(0, 0)

        self.assertRaises(TypeError, model.append, 42)
        self.assertRaises(TypeError, model.extend, [42])
        self.assertRaises(TypeError, model.insert, 0, 42)
        self.assertRaises(TypeError, model.remove, 0)
        self.assertRaises(TypeError, model.pop)
        self.assertRaises(TypeError, model.clear)
        self.assertRaises(TypeError, model.reverse)
        self.assertRaises(TypeError, model.sort)
        with self.assertRaises(TypeError):
            model[0] = 1
        with self.assertRaises(TypeError):
            del model[0]

        self.assertFalse(model.setData(index, domain[0], Qt.EditRole))
        self.assertTrue(model.setData(index, "foo", Qt.ToolTipRole))

        self.assertFalse(model.setItemData(index, {Qt.EditRole: domain[0],
                                                   Qt.ToolTipRole: "foo"}))
        self.assertTrue(model.setItemData(index, {Qt.ToolTipRole: "foo"}))

        self.assertFalse(model.insertRows(0, 1))
        self.assertEqual(list(model), list(domain.variables))
        self.assertFalse(model.removeRows(0, 1))
        self.assertEqual(list(model), list(domain.variables))


class TestContinuousPalettesModel(GuiTest):
    def setUp(self):
        self.palette1, self.palette2 = \
            list(colorpalettes.ContinuousPalettes.values())[:2]

    def test_all_categories(self):
        model = ContinuousPalettesModel()
        shown = {palette.name for palette in model.items
                 if isinstance(palette, colorpalettes.Palette)}
        expected = {palette.name
                    for palette in colorpalettes.ContinuousPalettes.values()}
        self.assertEqual(expected, shown)

        shown = {name for name in model.items if isinstance(name, str)}
        expected = {palette.category
                    for palette in colorpalettes.ContinuousPalettes.values()}
        self.assertEqual(expected, shown)

    def test_category_selection(self):
        categories = ('Diverging', 'Linear')
        model = ContinuousPalettesModel(categories=categories)
        shown = {palette.name
                 for palette in model.items
                 if isinstance(palette, colorpalettes.Palette)}
        expected = {palette.name
                    for palette in colorpalettes.ContinuousPalettes.values()
                    if palette.category in categories}
        self.assertEqual(expected, shown)
        self.assertIn("Diverging", model.items)
        self.assertIn("Linear", model.items)

    def test_single_category(self):
        category = 'Diverging'
        model = ContinuousPalettesModel(categories=(category, ))
        shown = {palette.name
                 for palette in model.items
                 if isinstance(palette, colorpalettes.Palette)}
        expected = {palette.name
                    for palette in colorpalettes.ContinuousPalettes.values()
                    if palette.category == category}
        self.assertEqual(expected, shown)
        self.assertEqual(len(model.items), len(shown))

    def test_count(self):
        model = ContinuousPalettesModel()
        model.items = [self.palette1, self.palette1]
        self.assertEqual(model.rowCount(QModelIndex()), 2)
        self.assertEqual(model.columnCount(QModelIndex()), 1)

    def test_data(self):
        model = ContinuousPalettesModel()
        model.items = ["Palettes", self.palette1, self.palette2]
        data = model.data
        index = model.index

        self.assertEqual(data(index(0, 0), Qt.EditRole), "Palettes")
        self.assertEqual(data(index(1, 0), Qt.EditRole),
                         self.palette1.friendly_name)
        self.assertEqual(data(index(2, 0), Qt.EditRole),
                         self.palette2.friendly_name)

        self.assertEqual(data(index(0, 0), Qt.DisplayRole), "Palettes")
        self.assertEqual(data(index(1, 0), Qt.DisplayRole),
                         self.palette1.friendly_name)
        self.assertEqual(data(index(2, 0), Qt.DisplayRole),
                         self.palette2.friendly_name)

        self.assertIsNone(data(index(0, 0), Qt.DecorationRole))
        with patch.object(self.palette1, "color_strip") as color_strip:
            self.assertIs(data(index(1, 0), Qt.DecorationRole),
                          color_strip.return_value)
        with patch.object(self.palette2, "color_strip") as color_strip:
            self.assertIs(data(index(2, 0), Qt.DecorationRole),
                          color_strip.return_value)

        self.assertIsNone(data(index(0, 0), Qt.UserRole))
        self.assertIs(data(index(1, 0), Qt.UserRole), self.palette1)
        self.assertIs(data(index(2, 0), Qt.UserRole), self.palette2)

        self.assertIsNone(data(index(2, 0), Qt.FontRole))

    def test_select_flags(self):
        model = ContinuousPalettesModel()
        model.items = ["Palettes", self.palette1, self.palette2]
        self.assertFalse(model.flags(model.index(0, 0)) & Qt.ItemIsSelectable)
        self.assertTrue(model.flags(model.index(1, 0)) & Qt.ItemIsSelectable)
        self.assertTrue(model.flags(model.index(2, 0)) & Qt.ItemIsSelectable)

    def testIndexOf(self):
        model = ContinuousPalettesModel()
        model.items = ["Palettes", self.palette1, self.palette2]
        self.assertEqual(model.indexOf(self.palette1), 1)
        self.assertEqual(model.indexOf(self.palette1.name), 1)
        self.assertEqual(model.indexOf(self.palette1.friendly_name), 1)
        self.assertEqual(model.indexOf(self.palette2), 2)
        self.assertEqual(model.indexOf(self.palette2.name), 2)
        self.assertEqual(model.indexOf(self.palette2.friendly_name), 2)
        self.assertIsNone(model.indexOf(42))


class TestPyListModelTooltip(GuiTest):
    def test_tooltips_size(self):
        def data(i):
            return model.data(model.index(i, 0))

        def tip(i):
            return model.data(model.index(i, 0), Qt.ToolTipRole)

        # Not enough tooptips - return None
        model = PyListModelTooltip(["foo", "bar", "baz"], ["footip", "bartip"])
        self.assertEqual(data(1), "bar")
        self.assertEqual(data(2), "baz")
        self.assertIsNone(data(3))
        self.assertEqual(tip(1), "bartip")
        self.assertIsNone(tip(2))

        # No tooltips
        model = PyListModelTooltip(["foo", "bar", "baz"])
        self.assertIsNone(tip(1))
        self.assertIsNone(tip(2))

        # Too many tooltips
        model = PyListModelTooltip(["foo", "bar"], ["footip", "bartip", "btip"])
        self.assertEqual(data(0), "foo")
        self.assertEqual(data(1), "bar")
        self.assertIsNone(data(2))
        self.assertEqual(tip(1), "bartip")
        self.assertEqual(tip(2), "btip")

    def test_tooltip_arg(self):
        def tip(i):
            return model.data(model.index(i, 0), Qt.ToolTipRole)

        # Allow generators
        s = dict(a="ta", b="tb")
        model = PyListModelTooltip(s, s.values())
        self.assertEqual(tip(0), "ta")
        self.assertEqual(tip(1), "tb")

        # Basically backward compatibility; this behaviour diverges from
        # behaviour of data role
        s = []
        model = PyListModelTooltip(["foo"], s)
        self.assertIsNone(tip(0))

        s += ["footip"]
        self.assertEqual(tip(1), "footip")


class TestTableModel(unittest.TestCase, dbt):
    def setUpDB(self):
        # pylint: disable=attribute-defined-outside-init
        self.conn, self.iris = self.create_iris_sql_table()

    def tearDownDB(self):
        self.drop_iris_sql_table()

    @dbt.run_on(["postgres", "mssql"])
    def test_dense_data(self):
        table = SqlTable(self.conn, self.iris, inspect_values=True)
        if self.current_db == "mssql":
            # when loading data from mssql db, Sql widget returns Table (not SqlTable)
            table = Table(table)
        new_domain = Domain(table.domain.attributes[:-1], table.domain.attributes[-1])
        table = table.transform(new_domain)
        model = TableModel(table)

        self._dense_data(table, model.data, model.index)

    def test_local_dense_data(self):
        table = Table("iris.tab")
        model = TableModel(table)

        self._dense_data(table, model.data, model.index)

    def _dense_data(self, table, data, index):
        # Y: categorical
        self.assertEqual(table[0, 4], data(index(0, 0), Qt.DisplayRole))
        self.assertIsInstance(data(index(0, 0), Qt.DisplayRole), str)
        self.assertEqual(table[0, 4], data(index(0, 0), Qt.EditRole))
        self.assertIsInstance(data(index(0, 0), Qt.EditRole), Value)
        self.assertIsInstance(data(index(0, 0), Qt.BackgroundRole), QBrush)
        self.assertIsInstance(data(index(0, 0), Qt.ForegroundRole), QColor)
        self.assertEqual(table[0, 4], data(index(0, 0), TableModel.ValueRole))
        self.assertEqual(table[0, 4], data(index(0, 0), TableModel.ClassValueRole))
        self.assertEqual(table.domain[4], data(index(0, 0), TableModel.VariableRole))
        self.assertIsInstance(data(index(0, 0), TableModel.VariableStatsRole), BasicStats)

        # X: continuous
        self.assertEqual(table[0, 0], data(index(0, 1), Qt.DisplayRole))
        self.assertIsInstance(data(index(0, 1), Qt.DisplayRole), str)
        self.assertEqual(table[0, 0], data(index(0, 1), Qt.EditRole))
        self.assertIsInstance(data(index(0, 1), Qt.EditRole), Value)
        self.assertIsNone(data(index(0, 1), Qt.BackgroundRole))
        self.assertIsNone(data(index(0, 1), Qt.ForegroundRole))
        self.assertEqual(table[0, 0], data(index(0, 1), TableModel.ValueRole))
        self.assertEqual(table[0, 4], data(index(0, 1), TableModel.ClassValueRole))
        self.assertEqual(table.domain[0], data(index(0, 1), TableModel.VariableRole))
        self.assertIsInstance(data(index(0, 1), TableModel.VariableStatsRole), BasicStats)

    def test_sparse_data(self):
        table = Table(test_filename("datasets/iris_basket.basket"))
        model = TableModel(table)
        data, index = model.data, model.index

        # Y: 2d
        self.assertListEqual([table[0, 4], table[0, 5], table[0, 6]],
                             [data(index(0, i), Qt.DisplayRole) for i in range(3)])
        self.assertIsInstance(data(index(0, 0), Qt.DisplayRole), str)
        self.assertEqual(table[0, 4], data(index(0, 0), Qt.EditRole))
        self.assertIsInstance(data(index(0, 0), Qt.EditRole), Value)
        self.assertIsInstance(data(index(0, 0), Qt.BackgroundRole), QBrush)
        self.assertIsInstance(data(index(0, 0), Qt.ForegroundRole), QColor)
        self.assertEqual(table[0, 4], data(index(0, 0), TableModel.ValueRole))
        self.assertIsNone(data(index(0, 0), TableModel.ClassValueRole))
        self.assertEqual(table.domain[4], data(index(0, 0), TableModel.VariableRole))
        self.assertIsInstance(data(index(0, 0), TableModel.VariableStatsRole), BasicStats)

        # X: sparse
        self.assertEqual("sepal_length=1.5, sepal_width=5.3, petal_length=4.1, petal_width=2",
                         data(index(0, 3), Qt.DisplayRole))
        self.assertIsNone(data(index(0, 3), Qt.EditRole))
        self.assertIsNone(data(index(0, 3), Qt.BackgroundRole))
        self.assertIsNone(data(index(0, 3), Qt.ForegroundRole))
        self.assertIsNone(data(index(0, 3), TableModel.ValueRole))
        self.assertIsNone(data(index(0, 3), TableModel.ClassValueRole))
        self.assertIsNone(data(index(0, 3), TableModel.VariableRole))
        self.assertIsNone(data(index(0, 3), TableModel.VariableStatsRole))

        # X: sparse_bool
        table = Table.from_numpy(Domain(table.domain.class_vars, metas=table.domain.attributes),
                                 csr_matrix(table.Y), metas=table.X)
        model = TableModel(table)
        data, index = model.data, model.index

        self.assertEqual("Iris-setosa", data(index(0, 1), Qt.DisplayRole))

        # metas: sparse
        self.assertEqual("sepal_length=1.5, sepal_width=5.3, petal_length=4.1, petal_width=2",
                         data(index(0, 0), Qt.DisplayRole))
        self.assertIsInstance(data(index(0, 0), Qt.BackgroundRole), QBrush)
        self.assertIsInstance(data(index(0, 0), Qt.ForegroundRole), QColor)


if __name__ == "__main__":
    unittest.main()



import sklearn.ensemble as skl_ensemble

from Orange.base import SklLearner
from Orange.classification.base_classification import (
    SklLearnerClassification, SklModelClassification
)
from Orange.regression.base_regression import (
    SklLearnerRegression, SklModelRegression
)

__all__ = ['SklAdaBoostClassificationLearner', 'SklAdaBoostRegressionLearner']


class SklAdaBoostClassifier(SklModelClassification):
    pass


class SklAdaBoostClassificationLearner(SklLearnerClassification):
    __wraps__ = skl_ensemble.AdaBoostClassifier
    __returns__ = SklAdaBoostClassifier
    supports_weights = True

    def __init__(self, base_estimator=None, n_estimators=50, learning_rate=1.,
                 algorithm='SAMME.R', random_state=None, preprocessors=None):
        from Orange.modelling import Fitter
        # If fitter, get the appropriate Learner instance
        if isinstance(base_estimator, Fitter):
            base_estimator = base_estimator.get_learner(
                base_estimator.CLASSIFICATION)
        # If sklearn learner, get the underlying sklearn representation
        if isinstance(base_estimator, SklLearner):
            base_estimator = base_estimator.__wraps__(**base_estimator.params)
        super().__init__(preprocessors=preprocessors)
        self.params = vars()


class SklAdaBoostRegressor(SklModelRegression):
    pass


class SklAdaBoostRegressionLearner(SklLearnerRegression):
    __wraps__ = skl_ensemble.AdaBoostRegressor
    __returns__ = SklAdaBoostRegressor
    supports_weights = True

    def __init__(self, base_estimator=None, n_estimators=50, learning_rate=1.,
                 loss='linear', random_state=None, preprocessors=None):
        from Orange.modelling import Fitter
        # If fitter, get the appropriate Learner instance
        if isinstance(base_estimator, Fitter):
            base_estimator = base_estimator.get_learner(
                base_estimator.REGRESSION)
        # If sklearn learner, get the underlying sklearn representation
        if isinstance(base_estimator, SklLearner):
            base_estimator = base_estimator.__wraps__(**base_estimator.params)
        super().__init__(preprocessors=preprocessors)
        self.params = vars()



import unittest
from unittest.mock import patch, Mock
import datetime
from collections import namedtuple

import numpy as np

from AnyQt.QtCore import Qt

from orangecanvas.scheme.signalmanager import LazyValue
from orangewidget.utils.signals import summarize

from Orange.data import Table, Domain, StringVariable, ContinuousVariable, \
    DiscreteVariable, TimeVariable
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.utils.state_summary import format_summary_details, \
    format_multiple_summaries

VarDataPair = namedtuple('VarDataPair', ['variable', 'data'])

# Continuous variable variations
continuous_full = VarDataPair(
    ContinuousVariable('continuous_full'),
    np.array([0, 1, 2, 3, 4], dtype=float),
)
continuous_missing = VarDataPair(
    ContinuousVariable('continuous_missing'),
    np.array([0, 1, 2, np.nan, 4], dtype=float),
)

# Unordered discrete variable variations
rgb_full = VarDataPair(
    DiscreteVariable('rgb_full', values=('r', 'g', 'b')),
    np.array([0, 1, 1, 1, 2], dtype=float),
)
rgb_missing = VarDataPair(
    DiscreteVariable('rgb_missing', values=('r', 'g', 'b')),
    np.array([0, 1, 1, np.nan, 2], dtype=float),
)

# Ordered discrete variable variations
ints_full = VarDataPair(
    DiscreteVariable('ints_full', values=('2', '3', '4')),
    np.array([0, 1, 1, 1, 2], dtype=float),
)
ints_missing = VarDataPair(
    DiscreteVariable('ints_missing', values=('2', '3', '4')),
    np.array([0, 1, 1, np.nan, 2], dtype=float),
)


def _to_timestamps(years):
    return [datetime.datetime(year, 1, 1).timestamp() if not np.isnan(year)
            else np.nan for year in years]


time_full = VarDataPair(
    TimeVariable('time_full'),
    np.array(_to_timestamps([2000, 2001, 2002, 2003, 2004]), dtype=float),
)
time_missing = VarDataPair(
    TimeVariable('time_missing'),
    np.array(_to_timestamps([2000, np.nan, 2001, 2003, 2004]), dtype=float),
)

# String variable variations
string_full = VarDataPair(
    StringVariable('string_full'),
    np.array(['a', 'b', 'c', 'd', 'e'], dtype=object),
)
string_missing = VarDataPair(
    StringVariable('string_missing'),
    np.array(['a', 'b', 'c', StringVariable.Unknown, 'e'], dtype=object),
)


def make_table(attributes, target=None, metas=None):
    """Build an instance of a table given various variables.

    Parameters
    ----------
    attributes : Iterable[Tuple[Variable, np.array]
    target : Optional[Iterable[Tuple[Variable, np.array]]
    metas : Optional[Iterable[Tuple[Variable, np.array]]

    Returns
    -------
    Table

    """
    attribute_vars, attribute_vals = list(zip(*attributes))
    attribute_vals = np.array(attribute_vals).T

    target_vars, target_vals = None, None
    if target is not None:
        target_vars, target_vals = list(zip(*target))
        target_vals = np.array(target_vals).T

    meta_vars, meta_vals = None, None
    if metas is not None:
        meta_vars, meta_vals = list(zip(*metas))
        meta_vals = np.array(meta_vals).T

    return Table.from_numpy(
        Domain(attribute_vars, class_vars=target_vars, metas=meta_vars),
        X=attribute_vals, Y=target_vals, metas=meta_vals,
    )


class TestUtils(unittest.TestCase):
    @WidgetTest.skipNonEnglish
    def test_details(self):
        """Check if details part of the summary is formatted correctly"""
        data = Table('zoo')
        n_features = len(data.domain.variables) + len(data.domain.metas)
        details = f'zoo: {len(data)} instances, ' \
                  f'{n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} categorical ' \
                  f'(no missing values)\n' \
                  f'Target: categorical\n' \
                  f'Metas: string'
        self.assertEqual(details, format_summary_details(data))

        details = f'Table with {n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} categorical\n' \
                  f'Target: categorical\n' \
                  f'Metas: string'
        self.assertEqual(details, format_summary_details(data.domain))

        data = Table('housing')
        n_features = len(data.domain.variables) + len(data.domain.metas)
        details = f'housing: {len(data)} instances, ' \
                  f'{n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} numeric ' \
                  f'(no missing values)\n' \
                  f'Target: numeric'
        self.assertEqual(details, format_summary_details(data))

        data = Table('heart_disease')
        n_features = len(data.domain.variables) + len(data.domain.metas)
        details = f'heart_disease: {len(data)} instances, ' \
                  f'{n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} ' \
                  f'(7 categorical, 6 numeric) (0.2% missing values)\n' \
                  f'Target: categorical'
        self.assertEqual(details, format_summary_details(data))

        data = make_table(
            [continuous_full, continuous_missing],
            target=[rgb_full, rgb_missing], metas=[ints_full, ints_missing]
        )
        n_features = len(data.domain.variables) + len(data.domain.metas)
        details = f'Table with {len(data)} instances, ' \
                  f'{n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} numeric ' \
                  f'(10.0% missing values)\n' \
                  f'Target: {len(data.domain.class_vars)} categorical\n' \
                  f'Metas: {len(data.domain.metas)} categorical'
        self.assertEqual(details, format_summary_details(data))

        data = make_table(
            [continuous_full, time_full, ints_full, rgb_missing],
            target=[rgb_full, continuous_missing],
            metas=[string_full, string_missing]
        )
        n_features = len(data.domain.variables) + len(data.domain.metas)
        details = f'Table with {len(data)} instances, ' \
                  f'{n_features} variables\n' \
                  f'Features: {len(data.domain.attributes)} ' \
                  f'(2 categorical, 1 numeric, 1 time) (5.0% missing values)\n' \
                  f'Target: {len(data.domain.class_vars)} ' \
                  f'(1 categorical, 1 numeric)\n' \
                  f'Metas: {len(data.domain.metas)} string'
        self.assertEqual(details, format_summary_details(data))

        data = make_table([time_full, time_missing], target=[ints_missing],
                          metas=None)
        details = f'Table with {len(data)} instances, ' \
                  f'{len(data.domain.variables)} variables\n' \
                  f'Features: {len(data.domain.attributes)} time ' \
                  f'(10.0% missing values)\n' \
                  f'Target: categorical'
        self.assertEqual(details, format_summary_details(data))

        data = make_table([rgb_full, ints_full], target=None, metas=None)
        details = f'Table with {len(data)} instances, ' \
                  f'{len(data.domain.variables)} variables\n' \
                  f'Features: {len(data.domain.variables)} categorical ' \
                  f'(no missing values)\n' \
                  f'Target: —'
        self.assertEqual(details, format_summary_details(data))

        data = make_table([rgb_full], target=None, metas=None)
        details = f'Table with {len(data)} instances, ' \
                  f'{len(data.domain.variables)} variable\n' \
                  f'Features: categorical (no missing values)\n' \
                  f'Target: —'
        self.assertEqual(details, format_summary_details(data))

        data = Table.from_numpy(domain=None, X=np.random.random((10000, 1000)))
        details = f'Table with {len(data):n} instances, ' \
                  f'{len(data.domain.variables)} variables\n' \
                  f'Features: {len(data.domain.variables)} numeric\n' \
                  f'Target: —'
        with patch.object(Table, "get_nan_frequency_attribute") as mock:
            self.assertEqual(details, format_summary_details(data))
            mock.assert_not_called()

        data = None
        self.assertEqual('', format_summary_details(data))

    @WidgetTest.skipNonEnglish
    def test_multiple_summaries(self):
        data = Table('zoo')
        extra_data = Table('zoo')[20:]
        n_features_data = len(data.domain.variables) + len(data.domain.metas)
        n_features_extra_data = len(extra_data.domain.variables) + \
                                len(extra_data.domain.metas)
        details = f'Data:<br>zoo: {len(data)} instances, ' \
                  f'{n_features_data} variables<br>' \
                  f'Features: {len(data.domain.attributes)} categorical ' \
                  f'(no missing values)<br>' \
                  f'Target: categorical<br>' \
                  f'Metas: string<hr>'\
                  f'Extra Data:<br>zoo: {len(extra_data)} instances, ' \
                  f'{n_features_extra_data} variables<br>' \
                  f'Features: {len(extra_data.domain.attributes)} ' \
                  f'categorical (no missing values)<br>' \
                  f'Target: categorical<br>' \
                  f'Metas: string'
        inputs = [('Data', data), ('Extra Data', extra_data)]
        self.assertEqual(details, format_multiple_summaries(inputs))

        details = f'zoo: {len(data)} instances, ' \
                  f'{n_features_data} variables<br>' \
                  f'Features: {len(data.domain.attributes)} categorical ' \
                  f'(no missing values)<br>' \
                  f'Target: categorical<br>' \
                  f'Metas: string<hr>'\
                  f'zoo: {len(extra_data)} instances, ' \
                  f'{n_features_extra_data} variables<br>' \
                  f'Features: {len(extra_data.domain.attributes)} ' \
                  f'categorical (no missing values)<br>' \
                  f'Target: categorical<br>' \
                  f'Metas: string'
        inputs = [('', data), ('', extra_data)]
        self.assertEqual(details, format_multiple_summaries(inputs))

        details = f'No data on output.<hr>' \
                  f'Extra data:<br>zoo: {len(extra_data)} instances, ' \
                  f'{n_features_extra_data} variables<br>' \
                  f'Features: {len(extra_data.domain.attributes)} ' \
                  f'categorical (no missing values)<br>' \
                  f'Target: categorical<br>' \
                  f'Metas: string<hr>'\
                  f'No data on output.'
        outputs = [('', None), ('Extra data', extra_data), ('', None)]
        self.assertEqual(details,
                         format_multiple_summaries(outputs, type_io='output'))


class TestSummarize(unittest.TestCase):
    @patch("Orange.widgets.utils.state_summary._table_previewer")
    def test_summarize_table(self, previewer):
        data = Table('zoo')
        summary = summarize(data)
        self.assertEqual(summary.summary, len(data))
        self.assertEqual(summary.details,
                         format_summary_details(data, format=Qt.RichText))
        previewer.assert_not_called()
        summary.preview_func()
        previewer.assert_called_with(data)

    @patch("Orange.widgets.utils.state_summary._table_previewer")
    def test_summarize_lazy_table(self, previewer):
        data = Table('zoo')

        # lazy_data of unknown length and domain
        lazy_data = LazyValue[Table](lambda: data)
        lazy_data.get_value = Mock(return_value=data)
        summary = summarize(lazy_data)
        self.assertEqual(summary.summary, "?")
        self.assertIsInstance(summary.details, str)
        lazy_data.get_value.assert_not_called()
        previewer.assert_not_called()
        summary.preview_func()
        lazy_data.get_value.assert_called()
        previewer.assert_called_with(data)
        previewer.reset_mock()

        # lazy_data with length and domain hint
        lazy_data = LazyValue[Table](
            lambda: data, length=123, domain=data.domain)
        lazy_data.get_value = Mock(return_value=data)
        summary = summarize(lazy_data)
        self.assertEqual(summary.summary, 123)
        self.assertEqual(summary.details,
                         format_summary_details(data.domain, format=Qt.RichText))
        lazy_data.get_value.assert_not_called()
        previewer.assert_not_called()
        summary.preview_func()
        lazy_data.get_value.assert_called()
        previewer.assert_called_with(data)
        previewer.reset_mock()

        # lazy_data that is already cached: complete summary even without hints
        lazy_data = LazyValue[Table](lambda: data)
        lazy_data.get_value()
        summary = summarize(lazy_data)
        self.assertEqual(summary.summary, len(data))
        self.assertEqual(summary.details,
                         format_summary_details(data, format=Qt.RichText))
        previewer.assert_not_called()
        summary.preview_func()
        previewer.assert_called_with(data)


if __name__ == "__main__":
    unittest.main()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring
from scipy.sparse import csr_matrix

from AnyQt.QtCore import Qt
from Orange.widgets.model.owsvm import OWSVM
from Orange.widgets.tests.base import (
    WidgetTest,
    DefaultParameterMapping,
    ParameterMapping,
    WidgetLearnerTestMixin
)
from Orange.data import Table


class TestOWSVMClassification(WidgetTest, WidgetLearnerTestMixin):
    def setUp(self):
        self.widget = self.create_widget(
            OWSVM, stored_settings={"auto_apply": False})
        self.init()

        # All this nonsense is necessary because we add an `auto` option to the
        # gamma spin box
        gamma_spin = self.widget._kernel_params[0]
        values = [self.widget._default_gamma, gamma_spin.maximum()]

        def getter():
            value = gamma_spin.value()
            return gamma_spin.specialValueText() \
                if value == gamma_spin.minimum() else value

        def setter(value):
            if value == gamma_spin.specialValueText():
                gamma_spin.setValue(gamma_spin.minimum())
            else:
                gamma_spin.setValue(value)

        self.parameters = [
            ParameterMapping("C", self.widget.c_spin),
            ParameterMapping("gamma", self.widget._kernel_params[0],
                             values=values, setter=setter, getter=getter),
            ParameterMapping("coef0", self.widget._kernel_params[1]),
            ParameterMapping("degree", self.widget._kernel_params[2]),
            ParameterMapping("tol", self.widget.tol_spin),
            ParameterMapping("max_iter", self.widget.max_iter_spin[1])]

    def test_parameters_unchecked(self):
        """Check learner and model for various values of all parameters
        when Iteration limit is not checked
        """
        self.widget.max_iter_spin[0].setCheckState(Qt.Unchecked)
        self.parameters[-1] = DefaultParameterMapping("max_iter", -1)
        self.test_parameters()

    def test_parameters_svm_type(self):
        """Check learner and model for various values of all parameters
        when NuSVM is chosen
        """
        self.assertEqual(self.widget.svm_type, OWSVM.SVM)
        # setChecked(True) does not trigger callback event
        self.widget.nu_radio.click()
        self.assertEqual(self.widget.svm_type, OWSVM.Nu_SVM)
        self.parameters[0] = ParameterMapping("nu", self.widget.nu_spin)
        self.test_parameters()

    def test_kernel_equation(self):
        """Check if the right equation is written according to kernel """
        for i in range(4):
            if self.widget.kernel_box.buttons[i].isChecked():
                self.assertEqual(self.widget.kernel_eq,
                                 self.widget.kernels[i][1])
                break
        for i in range(4):
            self.widget.kernel_box.buttons[i].click()
            self.assertEqual(self.widget.kernel_eq, self.widget.kernels[i][1])

    def test_kernel_spins(self):
        """Check if the right spins are visible according to kernel """
        for i, hidden in enumerate([[True, True, True],
                                    [False, False, False],
                                    [False, True, True],
                                    [False, False, True]]):
            if self.widget.kernel_box.buttons[i].isChecked():
                self.assertEqual([self.widget._kernel_params[j].box.isHidden()
                                  for j in range(3)], hidden)
                break
        for i, hidden in enumerate([[True, True, True],
                                    [False, False, False],
                                    [False, True, True],
                                    [False, False, True]]):
            self.widget.kernel_box.buttons[i].click()
            self.assertEqual([self.widget._kernel_params[j].box.isHidden()
                              for j in range(3)], hidden)

    def test_sparse_warning(self):
        """Check if the user is warned about sparse input"""
        data = Table("iris")
        self.send_signal(self.widget.Inputs.data, data)
        self.assertFalse(self.widget.Warning.sparse_data.is_shown())

        with data.unlocked():
            data.X = csr_matrix(data.X)
        self.send_signal(self.widget.Inputs.data, data)
        self.assertTrue(self.widget.Warning.sparse_data.is_shown())



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring, protected-access

import unittest
from unittest.mock import patch

import numpy as np

from AnyQt.QtCore import Qt

from Orange.data import (Table, Domain, StringVariable,
                         ContinuousVariable)
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin
from Orange.widgets.utils.annotated_data import (ANNOTATED_DATA_FEATURE_NAME)
from Orange.widgets.visualize.owvenndiagram import (
    OWVennDiagram, get_perm, arrays_equal, pad_columns,
    IDENTITY_STR, EQUALITY_STR)


class TestOWVennDiagram(WidgetTest, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls, False)

        cls.signal_data = cls.data[:25]

    def setUp(self):
        self.widget = self.create_widget(OWVennDiagram)
        self.signal_name = self.widget.Inputs.data

    def _select_data(self):
        self.widget.vennwidget.vennareas()[1].setSelected(True)
        return list(range(len(self.signal_data)))

    def test_rows_id(self):
        data = Table('zoo')
        data1 = data.copy()
        with data1.unlocked():
            data1[:, 1] = 1
        self.widget.rowwise = True
        self.send_signal(self.signal_name, data1[:10], 1)
        self.widget.selected_feature = IDENTITY_STR
        self.send_signal(self.signal_name, data[5:10], 2)

        self.assertIsNone(self.get_output(self.widget.Outputs.selected_data))
        self.assertTrue(self.widget.Warning.renamed_vars.is_shown())

        self.widget.vennwidget.vennareas()[3].setSelected(True)
        selected = self.get_output(self.widget.Outputs.selected_data)
        sel_atrs = [atr.name for atr in selected.domain.attributes]
        true_atrs = ['hair', 'feathers (1)', 'feathers (2)', 'eggs', 'milk', 'airborne', 'aquatic',
                     'predator', 'toothed', 'backbone', 'breathes', 'venomous', 'fins', 'legs',
                     'tail', 'domestic', 'catsize']
        self.assertEqual(sel_atrs, true_atrs)
        self.assertEqual(selected.domain.metas, data.domain.metas)
        self.assertEqual(selected.domain.class_vars, data.domain.class_vars)

    def test_output_duplicates(self):
        self.widget.rowwise = True
        self.widget.output_duplicates = True
        self.send_signal(self.signal_name, self.data[:2], 1)
        self.send_signal(self.signal_name, self.data[:4], 2)

        self.widget.vennwidget.vennareas()[3].setSelected(True)
        selected = self.get_output(self.widget.Outputs.selected_data)
        in_ids = self.data[:2].ids
        true_ids = np.array([in_ids[0], in_ids[0], in_ids[1], in_ids[1]])
        np.testing.assert_array_equal(selected.ids.flatten(), true_ids)

    def test_multiple_input_rows_id(self):
        """Over rows"""
        self.widget.rowwise = True
        self.send_signal(self.signal_name, self.data[:100], 1)
        self.send_signal(self.signal_name, self.data[50:], 2)

        # check selected data output
        self.assertIsNone(self.get_output(self.widget.Outputs.selected_data))

        # check annotated data output
        feature_name = ANNOTATED_DATA_FEATURE_NAME
        annotated = self.get_output(self.widget.Outputs.annotated_data)
        self.assertEqual(0, np.sum([i[feature_name] for i in annotated]))

        # select data instances
        self.widget.vennwidget.vennareas()[3].setSelected(True)
        selected_indices = list(range(50, 100))

        # check selected data output
        selected = self.get_output(self.widget.Outputs.selected_data)
        n_sel, n_attr = len(selected), len(self.data.domain.attributes)
        self.assertGreater(n_sel, 0)
        self.assertEqual(selected.domain == self.data.domain,
                         self.same_input_output_domain)
        np.testing.assert_array_equal(selected.X[:, :n_attr],
                                      self.data.X[selected_indices])

        # check annotated data output
        annotated = self.get_output(self.widget.Outputs.annotated_data)
        self.assertEqual(n_sel, np.sum([i[feature_name] for i in annotated]))

        # compare selected and annotated data domains
        self._compare_selected_annotated_domains(selected, annotated)

        # check output when data is removed
        self.send_signal(self.signal_name, None, 1)
        self.send_signal(self.signal_name, None, 2)
        self.assertIsNone(self.get_output(self.widget.Outputs.selected_data))
        self.assertIsNone(self.get_output(self.widget.Outputs.annotated_data))

    def test_equality(self):
        self.widget.rowwise = True
        data1 = Table.from_numpy(None, np.arange(20).reshape(20, 1))
        data2 = Table.from_numpy(None, np.arange(10, 25).reshape(15, 1))
        self.send_signal(self.signal_name, data1, 1)
        self.send_signal(self.signal_name, data2, 2)

        self.widget.vennwidget.vennareas()[3].setSelected(True)
        self.assertFalse(bool(self.get_output(self.widget.Outputs.selected_data)))

        self.widget.selected_feature = EQUALITY_STR
        self.widget._on_inputAttrActivated()
        self.assertEqual(
            set(self.get_output(self.widget.Outputs.selected_data).X.flatten()),
            set(range(10, 20)))

    def test_disable_duplicates(self):
        cb = self.widget.output_duplicates_cb
        data1 = Table.from_numpy(None, np.arange(20).reshape(20, 1))
        data2 = Table.from_numpy(None, np.arange(10, 25).reshape(15, 1))
        self.send_signal(self.signal_name, data1, 1)
        self.send_signal(self.signal_name, data2, 2)
        self.assertFalse(cb.isEnabled())

        data = Table("zoo")
        self.send_signal(self.signal_name, data, 1)
        self.send_signal(self.signal_name, data, 2)
        for self.widget.selected_feature in (IDENTITY_STR, EQUALITY_STR):
            self.widget._on_inputAttrActivated()
            self.assertFalse(cb.isEnabled())

        self.widget.selected_feature = \
            self.widget.controls.selected_feature.model()[2]
        self.widget._on_inputAttrActivated()
        self.assertTrue(cb.isEnabled())

        self.widget.rowwise = False
        self.widget._on_matching_changed()
        self.assertFalse(cb.isEnabled())

    def test_disable_match_equality(self):
        model = self.widget.controls.selected_feature.model()
        row0 = model.index(0, 0)
        row1 = model.index(1, 0)
        data = Table("zoo")

        self.assertTrue(model.flags(row0) & Qt.ItemIsEnabled)
        self.assertTrue(model.flags(row1) & Qt.ItemIsEnabled)

        self.send_signal(self.signal_name, data, 1)
        self.assertTrue(model.flags(row0) & Qt.ItemIsEnabled)
        self.assertTrue(model.flags(row1) & Qt.ItemIsEnabled)

        self.send_signal(self.signal_name, data, 2)
        self.assertTrue(model.flags(row0) & Qt.ItemIsEnabled)
        self.assertTrue(model.flags(row1) & Qt.ItemIsEnabled)

        data1 = Table.from_numpy(None, np.arange(20).reshape(20, 1))
        self.widget.selected_feature = EQUALITY_STR
        self.send_signal(self.signal_name, data1, 3)
        self.assertEqual(self.widget.selected_feature, IDENTITY_STR)
        self.assertTrue(model.flags(row0) & Qt.ItemIsEnabled)
        self.assertFalse(model.flags(row1) & Qt.ItemIsEnabled)

    def test_multiple_input_over_cols(self):
        self.widget.rowwise = False
        selected_atr_name = 'Selected'
        input2 = self.data.transform(Domain([self.data.domain.attributes[0]],
                                            self.data.domain.class_vars,
                                            self.data.domain.metas)).copy()
        self.send_signal(self.signal_name, self.data, (1, 'Data', None))
        self.send_signal(self.signal_name, input2, (2, 'Data', None))

        selected = self.get_output(self.widget.Outputs.selected_data)
        annotated = self.get_output(self.widget.Outputs.annotated_data)
        self.assertIsNone(selected)
        self.assertEqual(len(annotated), len(self.data))
        self.assertEqual(annotated.domain, self.data.domain)
        for atr in annotated.domain.attributes:
            self.assertFalse(atr.attributes)

        # select data instances
        self.widget.vennwidget.vennareas()[3].setSelected(True)
        selected = self.get_output(self.widget.Outputs.selected_data)
        np.testing.assert_array_equal(selected.X,
                                      input2.X)
        np.testing.assert_array_equal(selected.Y,
                                      input2.Y)
        np.testing.assert_array_equal(selected.metas,
                                      input2.metas)

        #domain matches but the values do not
        with input2.unlocked(input2.X):
            input2.X = input2.X - 1
        self.send_signal(self.signal_name, input2, (2, 'Data', None))
        self.widget.vennwidget.vennareas()[3].setSelected(True)
        annotated = self.get_output(self.widget.Outputs.annotated_data)
        selected = self.get_output(self.widget.Outputs.selected_data)
        atrs = {atr.name for atr in selected.domain.attributes}
        true_atrs = {'sepal length (2)', 'sepal length (1)'}
        self.assertTrue(atrs == true_atrs)

        out_domain = annotated.domain.attributes
        self.assertTrue(out_domain[0].attributes[selected_atr_name])
        self.assertTrue(out_domain[1].attributes[selected_atr_name])
        self.assertFalse(out_domain[2].attributes[selected_atr_name])
        self.assertFalse(out_domain[3].attributes[selected_atr_name])
        self.assertFalse(out_domain[4].attributes[selected_atr_name])

    def test_test_explicit_closing(self):
        data = self.data[:3]
        self.widget.rowwise = True
        self.send_signal(self.signal_name, data[:1], 1)
        self.send_signal(self.signal_name, data[1:2], 2)
        self.send_signal(self.signal_name, data[2:3], 3)
        out = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(out.ids, data[:3].ids)

        self.send_signal(self.signal_name, None, 2)
        out = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(out.ids, data[0:3:2].ids)

        self.send_signal(self.signal_name, data[1:2], 2)
        out = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(out.ids, data[:3].ids)

        self.send_signal(self.signal_name,
                         self.widget.Inputs.data.closing_sentinel, 1)
        out = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(out.ids, data[1:3].ids)

    def test_no_data(self):
        """Check that the widget doesn't crash on empty data"""
        self.send_signal(self.signal_name, self.data[:0], 1)
        self.send_signal(self.signal_name, self.data[:100], 2)
        self.send_signal(self.signal_name, self.data[50:], 3)

        for i in range(1, 4):
            self.send_signal(self.signal_name, None, i)

        self.send_signal(self.signal_name, self.data[:100], 1)
        self.send_signal(self.signal_name, self.data[:0], 1)
        self.send_signal(self.signal_name, self.data[50:], 3)

        for i in range(1, 4):
            self.send_signal(self.signal_name, None, i)

        self.send_signal(self.signal_name, self.data[:100], 1)
        self.send_signal(self.signal_name, self.data[50:], 2)
        self.send_signal(self.signal_name, self.data[:0], 3)

    def test_unconditional_commit_on_new_signal(self):
        with patch.object(self.widget.commit, 'now') as commit:
            self.widget.autocommit = False
            commit.reset_mock()
            self.send_signal(self.signal_name, self.data[:100], 1)
            commit.assert_called()

    def test_input_compatibility(self):
        self.widget.rowwise = True
        self.send_signal(self.signal_name, self.data, 1)
        self.send_signal(self.signal_name,
                         self.data.transform(Domain([self.data.domain.attributes[0]],
                                                    self.data.domain.class_vars,
                                                    self.data.domain.metas)), 2)
        self.assertFalse(self.widget.Error.instances_mismatch.is_shown())

        self.widget.rowwise = False
        self.send_signal(self.signal_name, self.data[:100, :], 2)
        self.assertTrue(self.widget.Error.instances_mismatch.is_shown())

        self.send_signal(self.signal_name, None, 2)
        self.assertFalse(self.widget.Error.instances_mismatch.is_shown())

    def test_rows_identifiers(self):
        self.widget.rowwise = True
        data = Table('zoo')
        self.send_signal(self.signal_name, data, (1, 'Data', None))
        self.widget.selected_feature = data.domain.metas[0]
        self.send_signal(self.signal_name, data[:5], (2, 'Data', None))

        self.assertIsNone(self.get_output(self.widget.Outputs.selected_data))
        self.widget.vennwidget.vennareas()[3].setSelected(True)
        selected = self.get_output(self.widget.Outputs.selected_data)
        self.assertEqual(len(selected), 5)
        self.assertEqual(selected.domain.attributes, data.domain.attributes)
        self.assertEqual(selected.domain.class_vars, data.domain.class_vars)

        annotated = self.get_output(self.widget.Outputs.annotated_data)
        self.assertEqual(len(annotated), 100)

    def test_too_many_inputs(self):
        self.send_signal(self.signal_name, self.data, 1)
        self.send_signal(self.signal_name, self.data, 2)
        self.send_signal(self.signal_name, self.data, 3)
        self.send_signal(self.signal_name, self.data, 4)
        self.send_signal(self.signal_name, self.data, 5)
        self.send_signal(self.signal_name, self.data, 6)
        self.assertTrue(self.widget.Error.too_many_inputs.is_shown())

        self.send_signal(self.signal_name,
                         self.widget.Inputs.data.closing_sentinel, 6)
        self.assertFalse(self.widget.Error.too_many_inputs.is_shown())

    def test_no_attributes(self):
        domain = Domain([], class_vars=self.data.domain.attributes)
        n = len(self.data)
        table = Table.from_numpy(domain, np.empty((n, 0)), self.data.X)

        self.widget.rowwise = True
        self.send_signal(self.signal_name, table, 1)
        out = self.get_output(self.widget.Outputs.annotated_data)
        self.assertEqual(len(out), len(table))

    def test_migration_to_3(self):
        settings = {"selected_feature": None}
        OWVennDiagram.migrate_settings(settings, 1)
        self.assertEqual(settings["selected_feature"], IDENTITY_STR)

        settings = {}
        OWVennDiagram.migrate_settings(settings, 1)
        self.assertEqual(settings["selected_feature"], IDENTITY_STR)


class TestVennUtilities(unittest.TestCase):
    def test_array_equals_cols(self):
        a = np.array([1, 2], dtype=np.float64)
        b = np.array([1, np.nan], dtype=np.float64)
        self.assertTrue(arrays_equal(None, None, None))
        self.assertFalse(arrays_equal(None, a, None))
        self.assertFalse(arrays_equal(a, None, None))
        self.assertFalse(arrays_equal(a, b, ContinuousVariable))
        a[1] = np.nan
        self.assertTrue(arrays_equal(a, b, ContinuousVariable))
        self.assertTrue(arrays_equal(a.astype(str), a.astype(str), StringVariable))
        a[1] = 2
        b[1] = 3
        self.assertFalse(arrays_equal(a, b, ContinuousVariable))
        self.assertFalse(arrays_equal(a.astype(str), b.astype(str), StringVariable))

    def test_pad_columns(self):
        l = 5
        mask = [2, 3]
        values = np.array([7.2, 77.3]).reshape(-1, 1)
        res = pad_columns(values, mask, l)
        true_res = np.array([np.nan, np.nan, 7.2, 77.3, np.nan]).reshape(-1, 1)
        np.testing.assert_array_equal(res, true_res)

    def test_get_perm(self):
        all_ids = [1, 7, 22]
        res = get_perm([7, 33], all_ids)
        true_res = [1]
        self.assertEqual(res, true_res)

        res = get_perm([22, 1, 7], all_ids)
        true_res = [2, 0, 1]
        self.assertEqual(res, true_res)


if __name__ == "__main__":
    unittest.main()



import enum
from collections import OrderedDict
from datetime import datetime, timezone, timedelta

import numpy as np

from AnyQt.QtWidgets import (
    QWidget, QTableWidget, QHeaderView, QComboBox, QLineEdit, QToolButton,
    QMessageBox, QMenu, QListView, QGridLayout, QPushButton, QSizePolicy,
    QLabel, QDateTimeEdit)
from AnyQt.QtGui import (QDoubleValidator, QStandardItemModel, QStandardItem,
                         QFontMetrics, QPalette)
from AnyQt.QtCore import Qt, QPoint, QPersistentModelIndex, QLocale, \
    QDateTime, QDate, QTime

from orangewidget.utils.combobox import ComboBoxSearch
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.data import (
    ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable, Table)
import Orange.data.filter as data_filter
from Orange.data.filter import FilterContinuous, FilterString
from Orange.data.sql.table import SqlTable
from Orange.preprocess import Remove
from Orange.widgets import widget, gui
from Orange.widgets.settings import Setting, ContextSetting, DomainContextHandler
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.utils.localization import pl
from Orange.widgets.widget import Input, Output
from Orange.widgets.utils import vartype
from Orange.widgets import report
from Orange.widgets.widget import Msg
from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                 ANNOTATED_DATA_SIGNAL_NAME)


class SelectRowsContextHandler(DomainContextHandler):
    """Context handler that filters conditions"""

    # pylint: disable=arguments-differ
    def is_valid_item(self, setting, condition, attrs, metas):
        """Return True if condition applies to a variable in given domain."""
        varname, *_ = condition
        return varname in attrs or varname in metas

    def encode_setting(self, context, setting, value):
        if setting.name != 'conditions':
            return super().encode_settings(context, setting, value)

        encoded = []
        CONTINUOUS = vartype(ContinuousVariable("x"))
        for attr, op, values in value:
            if isinstance(attr, str):
                if OWSelectRows.AllTypes.get(attr) == CONTINUOUS:
                    values = [QLocale().toDouble(v)[0] for v in values]
                # None will match the value returned by all_vars.get
                encoded.append((attr, None, op, values))
            else:
                # check for exact match, pylint: disable=unidiomatic-typecheck
                if type(attr) is ContinuousVariable \
                        and values and isinstance(values[0], str):
                    values = [QLocale().toDouble(v)[0] for v in values]
                elif isinstance(attr, DiscreteVariable):
                    values = [attr.values[i - 1] if i else "" for i in values]
                encoded.append((
                    attr.name,
                    context.attributes.get(attr.name)
                    or context.metas.get(attr.name),
                    op,
                    values
                ))
        return encoded

    # pylint: disable=arguments-differ
    def decode_setting(self, setting, value, domain=None, *_args):
        value = super().decode_setting(setting, value, domain)
        if setting.name == 'conditions':
            CONTINUOUS = vartype(ContinuousVariable("x"))
            for i, (attr, tpe, op, values) in enumerate(value):
                if tpe is not None:
                    attr = domain[attr]
                # check for exact match, pylint: disable=unidiomatic-typecheck
                if type(attr) is ContinuousVariable \
                        or OWSelectRows.AllTypes.get(attr) == CONTINUOUS:
                    values = [QLocale().toString(float(i), 'f') for i in values]
                elif isinstance(attr, DiscreteVariable):
                    values = tuple(
                        attr.to_val(val) + 1 if val else 0
                        for val in values
                        if val in attr.values
                    ) or (0,)
                value[i] = (attr, op, values)
        return value

    def match(self, context, domain, attrs, metas):
        if (attrs, metas) == (context.attributes, context.metas):
            return self.PERFECT_MATCH

        conditions = context.values["conditions"]
        all_vars = attrs.copy()
        all_vars.update(metas)
        matched = [
            all_vars.get(name) == tpe  # also matches "all (...)" strings
            for name, tpe, *rest in conditions
        ]
        if any(matched):
            return 0.5 * sum(matched) / len(matched)
        return self.NO_MATCH

    def filter_value(self, setting, data, domain, attrs, metas):
        if setting.name != "conditions":
            super().filter_value(setting, data, domain, attrs, metas)
            return

        all_vars = attrs.copy()
        all_vars.update(metas)
        conditions = data["conditions"]
        conditions[:] = [
            (name, tpe, *rest)
            for name, tpe, *rest in conditions
            if all_vars.get(name) == tpe
        ]


class FilterDiscreteType(enum.Enum):
    # pylint: disable=invalid-name
    Equal = "Equal"
    NotEqual = "NotEqual"
    In = "In"
    IsDefined = "IsDefined"


class OWSelectRows(widget.OWWidget):
    name = "Select Rows"
    description = "Select rows from the data based on values of variables."
    icon = "icons/SelectRows.svg"
    priority = 100
    category = "Transform"
    keywords = "select rows, filter"

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        matching_data = Output("Matching Data", Table, default=True)
        unmatched_data = Output("Unmatched Data", Table)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    want_main_area = False

    settingsHandler = SelectRowsContextHandler()
    conditions = ContextSetting([])
    update_on_change = Setting(True)
    purge_attributes = Setting(False, schema_only=True)
    purge_classes = Setting(False, schema_only=True)
    auto_commit = Setting(True)

    settings_version = 2

    Operators = {
        ContinuousVariable: [
            (FilterContinuous.Equal, "equals", "equal"),
            (FilterContinuous.NotEqual, "is not", "are not"),
            (FilterContinuous.Less, "is below", "are below"),
            (FilterContinuous.LessEqual, "is at most", "are at most"),
            (FilterContinuous.Greater, "is greater than", "are greater than"),
            (FilterContinuous.GreaterEqual, "is at least", "are at least"),
            (FilterContinuous.Between, "is between", "are between"),
            (FilterContinuous.Outside, "is outside", "are outside"),
            (FilterContinuous.IsDefined, "is defined", "are defined"),
        ],
        DiscreteVariable: [
            (FilterDiscreteType.Equal, "is"),
            (FilterDiscreteType.NotEqual, "is not"),
            (FilterDiscreteType.In, "is one of"),
            (FilterDiscreteType.IsDefined, "is defined")
        ],
        StringVariable: [
            (FilterString.Equal, "equals", "equal"),
            (FilterString.NotEqual, "is not", "are not"),
            (FilterString.Less, "is before", "are before"),
            (FilterString.LessEqual, "is equal or before", "are equal or before"),
            (FilterString.Greater, "is after", "are after"),
            (FilterString.GreaterEqual, "is equal or after", "are equal or after"),
            (FilterString.Between, "is between", "are between"),
            (FilterString.Outside, "is outside", "are outside"),
            (FilterString.Contains, "contains", "contain"),
            (FilterString.NotContain, "does not contain", "do not contain"),
            (FilterString.StartsWith, "begins with", "begin with"),
            (FilterString.NotStartsWith, "does not begin with", "do not begin with"),
            (FilterString.EndsWith, "ends with", "end with"),
            (FilterString.NotEndsWith, "does not end with", "do not end with"),
            (FilterString.IsDefined, "is defined", "are defined"),
            (FilterString.NotIsDefined, "is not defined", "are not defined"),
        ]
    }

    Operators[TimeVariable] = Operators[ContinuousVariable]

    AllTypes = {}
    for _all_name, _all_type, _all_ops in (
            ("All variables", 0,
             [(None, "are defined")]),
            ("All numeric variables", 2,
             [(v, t) for v, _, t in Operators[ContinuousVariable]]),
            ("All string variables", 3,
             [(v, t) for v, _, t in Operators[StringVariable]])):
        Operators[_all_name] = _all_ops
        AllTypes[_all_name] = _all_type

    operator_names = {vtype: [name for _, name, *_ in filters]
                      for vtype, filters in Operators.items()}

    class Error(widget.OWWidget.Error):
        parsing_error = Msg("{}")

    def __init__(self):
        super().__init__()

        self.old_purge_classes = True

        self.conditions = []
        self.last_output_conditions = None
        self.data = None
        self.data_desc = self.match_desc = self.nonmatch_desc = None
        self.variable_model = DomainModel(
            [list(self.AllTypes), DomainModel.Separator,
             DomainModel.CLASSES, DomainModel.ATTRIBUTES, DomainModel.METAS])

        box = gui.vBox(self.controlArea, 'Conditions', stretch=100)
        self.cond_list = QTableWidget(
            box, showGrid=False, selectionMode=QTableWidget.NoSelection)
        box.layout().addWidget(self.cond_list)
        self.cond_list.setColumnCount(4)
        self.cond_list.setRowCount(0)
        self.cond_list.verticalHeader().hide()
        self.cond_list.horizontalHeader().hide()
        for i in range(3):
            self.cond_list.horizontalHeader().setSectionResizeMode(i, QHeaderView.Stretch)
        self.cond_list.horizontalHeader().resizeSection(3, 30)
        self.cond_list.viewport().setBackgroundRole(QPalette.Window)

        box2 = gui.hBox(box)
        gui.rubber(box2)
        self.add_button = gui.button(
            box2, self, "Add Condition", callback=self.add_row)
        self.add_all_button = gui.button(
            box2, self, "Add All Variables", callback=self.add_all)
        self.remove_all_button = gui.button(
            box2, self, "Remove All", callback=self.remove_all)
        gui.rubber(box2)

        box_setting = gui.vBox(self.buttonsArea)
        self.cb_pa = gui.checkBox(
            box_setting, self, "purge_attributes",
            "Remove unused values and constant features",
            callback=self.conditions_changed)
        self.cb_pc = gui.checkBox(
            box_setting, self, "purge_classes", "Remove unused classes",
            callback=self.conditions_changed)

        self.report_button.setFixedWidth(120)
        gui.rubber(self.buttonsArea.layout())

        gui.auto_send(self.buttonsArea, self, "auto_commit")

        self.set_data(None)
        self.resize(600, 400)

    def add_row(self, attr=None, condition_type=None, condition_value=None):
        model = self.cond_list.model()
        row = model.rowCount()
        model.insertRow(row)

        attr_combo = ComboBoxSearch(
            minimumContentsLength=12,
            sizeAdjustPolicy=QComboBox.AdjustToMinimumContentsLengthWithIcon)
        attr_combo.setModel(self.variable_model)
        attr_combo.row = row
        attr_combo.setCurrentIndex(self.variable_model.indexOf(attr) if attr
                                   else len(self.AllTypes) + 1)
        self.cond_list.setCellWidget(row, 0, attr_combo)

        index = QPersistentModelIndex(model.index(row, 3))
        temp_button = QPushButton('×', self, flat=True,
                                  styleSheet='* {font-size: 16pt; color: palette(button-text) }'
                                             '*:hover {color: palette(bright-text)}')
        temp_button.clicked.connect(lambda: self.remove_one(index.row()))
        self.cond_list.setCellWidget(row, 3, temp_button)

        self.remove_all_button.setDisabled(False)
        self.set_new_operators(attr_combo, attr is not None,
                               condition_type, condition_value)
        attr_combo.currentIndexChanged.connect(
            lambda _: self.set_new_operators(attr_combo, False))

        self.cond_list.resizeRowToContents(row)

    def add_all(self):
        if self.cond_list.rowCount():
            Mb = QMessageBox
            if Mb.question(
                    self, "Remove existing filters",
                    "This will replace the existing filters with "
                    "filters for all variables.", Mb.Ok | Mb.Cancel) != Mb.Ok:
                return
            self.remove_all()
        for attr in self.variable_model[len(self.AllTypes) + 1:]:
            self.add_row(attr)
        self.conditions_changed()

    def remove_one(self, rownum):
        self.remove_one_row(rownum)
        self.conditions_changed()

    def remove_all(self):
        self.remove_all_rows()
        self.conditions_changed()

    def remove_one_row(self, rownum):
        self.cond_list.removeRow(rownum)
        if self.cond_list.model().rowCount() == 0:
            self.remove_all_button.setDisabled(True)

    def remove_all_rows(self):
        # Disconnect signals to avoid stray emits when changing variable_model
        for row in range(self.cond_list.rowCount()):
            for col in (0, 1):
                widg = self.cond_list.cellWidget(row, col)
                if widg:
                    widg.currentIndexChanged.disconnect()
        self.cond_list.clear()
        self.cond_list.setRowCount(0)
        self.remove_all_button.setDisabled(True)

    def set_new_operators(self, attr_combo, adding_all,
                          selected_index=None, selected_values=None):
        old_combo = self.cond_list.cellWidget(attr_combo.row, 1)
        prev_text = old_combo.currentText() if old_combo else ""
        oper_combo = QComboBox()
        oper_combo.row = attr_combo.row
        oper_combo.attr_combo = attr_combo
        attr_name = attr_combo.currentText()
        if attr_name in self.AllTypes:
            oper_combo.addItems(self.operator_names[attr_name])
        else:
            var = self.data.domain[attr_name]
            oper_combo.addItems(self.operator_names[type(var)])
        if selected_index is None:
            selected_index = oper_combo.findText(prev_text)
            if selected_index == -1:
                selected_index = 0
        oper_combo.setCurrentIndex(selected_index)
        self.cond_list.setCellWidget(oper_combo.row, 1, oper_combo)
        self.set_new_values(oper_combo, adding_all, selected_values)
        oper_combo.currentIndexChanged.connect(
            lambda _: self.set_new_values(oper_combo, False))

    @staticmethod
    def _get_lineedit_contents(box):
        contents = []
        for child in getattr(box, "controls", [box]):
            if isinstance(child, QLineEdit):
                contents.append(child.text())
            elif isinstance(child, DateTimeWidget):
                if child.format == (0, 1):
                    contents.append(child.time())
                elif child.format == (1, 0):
                    contents.append(child.date())
                elif child.format == (1, 1):
                    contents.append(child.dateTime())
        return contents

    @staticmethod
    def _get_value_contents(box):
        cont = []
        names = []
        for child in getattr(box, "controls", [box]):
            if isinstance(child, QLineEdit):
                cont.append(child.text())
            elif isinstance(child, QComboBox):
                cont.append(child.currentIndex())
            elif isinstance(child, QToolButton):
                if child.popup is not None:
                    model = child.popup.list_view.model()
                    for row in range(model.rowCount()):
                        item = model.item(row)
                        if item.checkState() == Qt.Checked:
                            cont.append(row + 1)
                            names.append(item.text())
                    child.desc_text = ', '.join(names)
                    child.set_text()
            elif isinstance(child, DateTimeWidget):
                if child.format == (0, 1):
                    cont.append(child.time())
                elif child.format == (1, 0):
                    cont.append(child.date())
                elif child.format == (1, 1):
                    cont.append(child.dateTime())
            elif isinstance(child, QLabel) or child is None:
                pass
            else:
                raise TypeError('Type %s not supported.' % type(child))
        return tuple(cont)

    class QDoubleValidatorEmpty(QDoubleValidator):
        def validate(self, input_, pos):
            if not input_:
                return QDoubleValidator.Acceptable, input_, pos
            if self.locale().groupSeparator() in input_:
                return QDoubleValidator.Invalid, input_, pos
            return super().validate(input_, pos)

    def set_new_values(self, oper_combo, adding_all, selected_values=None):
        # def remove_children():
        #     for child in box.children()[1:]:
        #         box.layout().removeWidget(child)
        #         child.setParent(None)

        def add_textual(contents):
            le = gui.lineEdit(box, self, None,
                              sizePolicy=QSizePolicy(QSizePolicy.Expanding,
                                                     QSizePolicy.Expanding))
            if contents:
                le.setText(contents)
            le.setAlignment(Qt.AlignRight)
            le.editingFinished.connect(self.conditions_changed)
            return le

        def add_numeric(contents):
            le = add_textual(contents)
            le.setValidator(OWSelectRows.QDoubleValidatorEmpty())
            return le

        box = self.cond_list.cellWidget(oper_combo.row, 2)
        lc = ["", ""]
        oper = oper_combo.currentIndex()
        attr_name = oper_combo.attr_combo.currentText()
        if attr_name in self.AllTypes:
            vtype = self.AllTypes[attr_name]
            var = None
        else:
            var = self.data.domain[attr_name]
            var_idx = self.data.domain.index(attr_name)
            vtype = vartype(var)
            if selected_values is not None:
                lc = list(selected_values) + ["", ""]
                lc = [str(x) if vtype != 4 else x for x in lc[:2]]
        if box and vtype == box.var_type:
            lc = self._get_lineedit_contents(box) + lc

        if "defined" in oper_combo.currentText():
            label = QLabel()
            label.var_type = vtype
            self.cond_list.setCellWidget(oper_combo.row, 2, label)
        elif var is not None and var.is_discrete:
            if oper_combo.currentText().endswith(" one of"):
                if selected_values:
                    lc = list(selected_values)
                button = DropDownToolButton(self, var, lc)
                button.var_type = vtype
                self.cond_list.setCellWidget(oper_combo.row, 2, button)
            else:
                combo = ComboBoxSearch()
                combo.addItems(("", ) + var.values)
                if lc[0]:
                    combo.setCurrentIndex(int(lc[0]))
                else:
                    combo.setCurrentIndex(0)
                combo.var_type = vartype(var)
                self.cond_list.setCellWidget(oper_combo.row, 2, combo)
                combo.currentIndexChanged.connect(self.conditions_changed)
        else:
            box = gui.hBox(self, addToLayout=False)
            box.var_type = vtype
            self.cond_list.setCellWidget(oper_combo.row, 2, box)
            if vtype == 2:  # continuous:
                box.controls = [add_numeric(lc[0])]
                if oper > 5:
                    gui.widgetLabel(box, " and ")
                    box.controls.append(add_numeric(lc[1]))
            elif vtype == 3:  # string:
                box.controls = [add_textual(lc[0])]
                if oper in [6, 7]:
                    gui.widgetLabel(box, " and ")
                    box.controls.append(add_textual(lc[1]))
            elif vtype == 4:  # time:
                def invalidate_datetime():
                    if w_:
                        if w.dateTime() > w_.dateTime():
                            w_.setDateTime(w.dateTime())
                        if w.format == (1, 1):
                            w.calendarWidget.timeedit.setTime(w.time())
                            w_.calendarWidget.timeedit.setTime(w_.time())
                    elif w.format == (1, 1):
                        w.calendarWidget.timeedit.setTime(w.time())

                def datetime_changed():
                    self.conditions_changed()
                    invalidate_datetime()

                datetime_format = (var.have_date, var.have_time)
                column = self.data.get_column(var_idx)
                w = DateTimeWidget(self, column, datetime_format)
                w.set_datetime(lc[0])
                box.controls = [w]
                box.layout().addWidget(w)
                w.dateTimeChanged.connect(datetime_changed)
                if oper > 5:
                    gui.widgetLabel(box, " and ")
                    w_ = DateTimeWidget(self, column, datetime_format)
                    w_.set_datetime(lc[1])
                    box.layout().addWidget(w_)
                    box.controls.append(w_)
                    invalidate_datetime()
                    w_.dateTimeChanged.connect(datetime_changed)
                else:
                    w_ = None
            else:
                box.controls = []
        if not adding_all:
            self.conditions_changed()

    @Inputs.data
    def set_data(self, data):
        self.closeContext()
        self.data = data
        self.cb_pa.setEnabled(not isinstance(data, SqlTable))
        self.cb_pc.setEnabled(not isinstance(data, SqlTable))
        self.remove_all_rows()
        self.add_button.setDisabled(data is None)
        self.add_all_button.setDisabled(
            data is None or
            len(data.domain.variables) + len(data.domain.metas) > 100)
        if not data:
            self.data_desc = None
            self.variable_model.set_domain(None)
            self.commit.deferred()
            return
        self.data_desc = report.describe_data_brief(data)
        self.variable_model.set_domain(data.domain)

        self.conditions = []
        self.openContext(data)
        for attr, cond_type, cond_value in self.conditions:
            if attr in self.variable_model:
                self.add_row(attr, cond_type, cond_value)
        if not self.cond_list.model().rowCount():
            self.add_row()

        self.commit.now()

    def conditions_changed(self):
        try:
            cells_by_rows = (
                [self.cond_list.cellWidget(row, col) for col in range(3)]
                for row in range(self.cond_list.rowCount())
            )
            self.conditions = [
                (var_cell.currentData(gui.TableVariable) or var_cell.currentText(),
                 oper_cell.currentIndex(),
                 self._get_value_contents(val_cell))
                for var_cell, oper_cell, val_cell in cells_by_rows]
            if self.update_on_change and (
                    self.last_output_conditions is None or
                    self.last_output_conditions != self.conditions):
                self.commit.deferred()
        except AttributeError:
            # Attribute error appears if the signal is triggered when the
            # controls are being constructed
            pass

    @staticmethod
    def _values_to_floats(attr, values):
        if len(values) == 0:
            return values
        if not all(values):
            return None
        if isinstance(attr, TimeVariable):
            values = (value.toString(format=Qt.ISODate) for value in values)
            parse = lambda x: (attr.parse(x), True)
        else:
            parse = QLocale().toDouble

        try:
            floats, ok = zip(*[parse(v) for v in values])
            if not all(ok):
                raise ValueError('Some values could not be parsed as floats'
                                 f' in the current locale: {values}')
        except TypeError:
            floats = values  # values already floats
        assert all(isinstance(v, float) for v in floats)
        return floats

    @gui.deferred
    def commit(self):
        matching_output = self.data
        non_matching_output = None
        annotated_output = None

        self.Error.clear()
        if self.data:
            domain = self.data.domain
            conditions = []
            for attr_name, oper_idx, values in self.conditions:
                if attr_name in self.AllTypes:
                    attr_index = attr = None
                    attr_type = self.AllTypes[attr_name]
                    operators = self.Operators[attr_name]
                else:
                    attr_index = domain.index(attr_name)
                    attr = domain[attr_index]
                    attr_type = vartype(attr)
                    operators = self.Operators[type(attr)]
                opertype, *_ = operators[oper_idx]
                if attr_type == 0:
                    filt = data_filter.IsDefined()
                elif attr_type in (2, 4):  # continuous, time
                    try:
                        floats = self._values_to_floats(attr, values)
                    except ValueError as e:
                        self.Error.parsing_error(e.args[0])
                        return
                    if floats is None:
                        continue
                    filt = data_filter.FilterContinuous(
                        attr_index, opertype, *floats)
                elif attr_type == 3:  # string
                    filt = data_filter.FilterString(
                        attr_index, opertype, *[str(v) for v in values])
                else:
                    if opertype == FilterDiscreteType.IsDefined:
                        f_values = None
                    else:
                        if not values or not values[0]:
                            continue
                        values = [attr.values[i-1] for i in values]
                        if opertype == FilterDiscreteType.Equal:
                            f_values = {values[0]}
                        elif opertype == FilterDiscreteType.NotEqual:
                            f_values = set(attr.values)
                            f_values.remove(values[0])
                        elif opertype == FilterDiscreteType.In:
                            f_values = set(values)
                        else:
                            raise ValueError("invalid operand")
                    filt = data_filter.FilterDiscrete(attr_index, f_values)
                conditions.append(filt)

            if conditions:
                filters = data_filter.Values(conditions)
                matching_output = filters(self.data)
                filters.negate = True
                non_matching_output = filters(self.data)

                row_sel = np.in1d(self.data.ids, matching_output.ids)
                annotated_output = create_annotated_table(self.data, row_sel)

            # if hasattr(self.data, "name"):
            #     matching_output.name = self.data.name
            #     non_matching_output.name = self.data.name

            purge_attrs = self.purge_attributes
            purge_classes = self.purge_classes
            if (purge_attrs or purge_classes) and \
                    not isinstance(self.data, SqlTable):
                attr_flags = sum([Remove.RemoveConstant * purge_attrs,
                                  Remove.RemoveUnusedValues * purge_attrs])
                class_flags = sum([Remove.RemoveConstant * purge_classes,
                                   Remove.RemoveUnusedValues * purge_classes])
                # same settings used for attributes and meta features
                remover = Remove(attr_flags, class_flags, attr_flags)

                matching_output = remover(matching_output)
                non_matching_output = remover(non_matching_output)
                annotated_output = remover(annotated_output)

        if not matching_output:
            matching_output = None
        if not non_matching_output:
            non_matching_output = None
        if not annotated_output:
            annotated_output = None

        self.Outputs.matching_data.send(matching_output)
        self.Outputs.unmatched_data.send(non_matching_output)
        self.Outputs.annotated_data.send(annotated_output)

        self.match_desc = report.describe_data_brief(matching_output)
        self.nonmatch_desc = report.describe_data_brief(non_matching_output)

    def send_report(self):
        if not self.data:
            self.report_paragraph("No data.")
            return

        pdesc = None
        describe_domain = False
        for d in (self.data_desc, self.match_desc, self.nonmatch_desc):
            if not d or not d["Data instances"]:
                continue
            ndesc = d.copy()
            del ndesc["Data instances"]
            if pdesc is not None and pdesc != ndesc:
                describe_domain = True
            pdesc = ndesc

        conditions = []
        for attr, oper, values in self.conditions:
            if isinstance(attr, str):
                attr_name = attr
                var_type = self.AllTypes[attr]
                names = self.operator_names[attr_name]
            else:
                attr_name = attr.name
                var_type = vartype(attr)
                names = self.operator_names[type(attr)]
            name = names[oper]
            if oper == len(names) - 1:
                conditions.append("{} {}".format(attr_name, name))
            elif var_type == 1:  # discrete
                if name == "is one of":
                    valnames = [attr.values[v - 1] for v in values]
                    if not valnames:
                        continue
                    if len(valnames) == 1:
                        valstr = valnames[0]
                    else:
                        valstr = f"{', '.join(valnames[:-1])} or {valnames[-1]}"
                    conditions.append(f"{attr} is {valstr}")
                elif values and values[0]:
                    value = values[0] - 1
                    conditions.append(f"{attr} {name} {attr.values[value]}")
            elif var_type == 3:  # string variable
                conditions.append(
                    f"{attr} {name} {' and '.join(map(repr, values))}")
            elif var_type == 4:  # time
                values = (value.toString(format=Qt.ISODate) for value in values)
                conditions.append(f"{attr} {name} {' and '.join(values)}")
            elif all(x for x in values):  # numeric variable
                conditions.append(f"{attr} {name} {' and '.join(values)}")
        items = OrderedDict()
        if describe_domain:
            items.update(self.data_desc)
        else:
            items["Instances"] = self.data_desc["Data instances"]
        items["Condition"] = " AND ".join(conditions) or "no conditions"
        self.report_items("Data", items)
        if describe_domain:
            self.report_items("Matching data", self.match_desc)
            self.report_items("Non-matching data", self.nonmatch_desc)
        else:
            match_inst = \
                bool(self.match_desc) and \
                self.match_desc["Data instances"]
            nonmatch_inst = \
                bool(self.nonmatch_desc) and \
                self.nonmatch_desc["Data instances"]
            self.report_items(
                "Output",
                (("Matching data",
                  f"{match_inst} {pl(match_inst, 'instance')}" if match_inst else "None"),
                 ("Non-matching data",
                  nonmatch_inst > 0 and f"{nonmatch_inst} {pl(nonmatch_inst, 'instance')}")))

    @classmethod
    def migrate_context(cls, context, version):
        if not version or version < 2:
            # Just remove; can't migrate because variables types are unknown
            context.values["conditions"] = []


class CheckBoxPopup(QWidget):
    def __init__(self, var, lc, widget_parent=None, widg=None):
        QWidget.__init__(self)

        self.list_view = QListView()
        text = []
        model = QStandardItemModel(self.list_view)
        for (i, val) in enumerate(var.values):
            item = QStandardItem(val)
            item.setCheckable(True)
            if i + 1 in lc:
                item.setCheckState(Qt.Checked)
                text.append(val)
            model.appendRow(item)
        model.itemChanged.connect(widget_parent.conditions_changed)
        self.list_view.setModel(model)

        layout = QGridLayout(self)
        layout.addWidget(self.list_view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.adjustSize()
        self.setWindowFlags(Qt.Popup)

        self.widget = widg
        self.widget.desc_text = ', '.join(text)
        self.widget.set_text()

    def moved(self):
        point = self.widget.rect().bottomRight()
        global_point = self.widget.mapToGlobal(point)
        self.move(global_point - QPoint(self.width(), 0))


class DropDownToolButton(QToolButton):
    def __init__(self, parent, var, lc):
        QToolButton.__init__(self, parent)
        self.desc_text = ''
        self.popup = CheckBoxPopup(var, lc, parent, self)
        self.setMenu(QMenu()) # to show arrow
        self.clicked.connect(self.open_popup)

    def open_popup(self):
        self.popup.moved()
        self.popup.show()

    def set_text(self):
        metrics = QFontMetrics(self.font())
        self.setText(metrics.elidedText(self.desc_text, Qt.ElideRight,
                                        self.width() - 15))

    def resizeEvent(self, _):
        self.set_text()


class DateTimeWidget(QDateTimeEdit):
    def __init__(self, parent, column, datetime_format):
        QDateTimeEdit.__init__(self, parent)

        self.format = datetime_format
        self.have_date, self.have_time = datetime_format[0], datetime_format[1]
        self.set_format(column)
        self.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))

    def set_format(self, column):
        str_format = Qt.ISODate
        if self.have_date and self.have_time:
            self.setDisplayFormat("yyyy-MM-dd hh:mm:ss")
            c_format = "%Y-%m-%d %H:%M:%S"
            min_datetime, max_datetime = self.find_range(column, c_format)
            self.min_datetime = QDateTime.fromString(min_datetime, str_format)
            self.max_datetime = QDateTime.fromString(max_datetime, str_format)
            self.setCalendarPopup(True)
            self.calendarWidget = gui.CalendarWidgetWithTime(
                self, time=self.min_datetime.time())
            self.calendarWidget.timeedit.timeChanged.connect(
                self.set_datetime)
            self.setCalendarWidget(self.calendarWidget)
            self.setDateTimeRange(self.min_datetime, self.max_datetime)

        elif self.have_date and not self.have_time:
            self.setDisplayFormat("yyyy-MM-dd")
            self.setCalendarPopup(True)
            min_datetime, max_datetime = self.find_range(column, "%Y-%m-%d")
            self.min_datetime = QDate.fromString(min_datetime, str_format)
            self.max_datetime = QDate.fromString(max_datetime, str_format)
            self.setDateRange(self.min_datetime, self.max_datetime)

        elif not self.have_date and self.have_time:
            self.setDisplayFormat("hh:mm:ss")
            min_datetime, max_datetime = self.find_range(column, "%H:%M:%S")
            self.min_datetime = QTime.fromString(min_datetime, str_format)
            self.max_datetime = QTime.fromString(max_datetime, str_format)
            self.setTimeRange(self.min_datetime, self.max_datetime)

    def set_datetime(self, date_time):
        if not date_time:
            date_time = self.min_datetime
        if self.have_date and self.have_time:
            if isinstance(date_time, QTime):
                self.setDateTime(
                    QDateTime(self.date(), self.calendarWidget.timeedit.time()))
            else:
                self.setDateTime(date_time)
        elif self.have_date and not self.have_time:
            self.setDate(date_time)
        elif not self.have_date and self.have_time:
            self.setTime(date_time)

    @staticmethod
    def find_range(column, convert_format):
        def convert_timestamp(timestamp):
            if timestamp >= 0:
                return datetime.fromtimestamp(timestamp, tz=timezone.utc)
            return datetime(1970, 1, 1, tzinfo=timezone.utc) + \
                       timedelta(seconds=int(timestamp))

        min_datetime = convert_timestamp(
            np.nanmin(column)).strftime(convert_format)
        max_datetime = convert_timestamp(
            np.nanmax(column)).strftime(convert_format)
        return min_datetime, max_datetime


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWSelectRows).run(Table("heart_disease"))



import numpy as np

from Orange.util import scale
from Orange.misc import DistMatrix
from Orange.widgets import widget, gui, settings
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output


class OWDistanceTransformation(widget.OWWidget):
    name = "Distance Transformation"
    description = "Transform distances according to selected criteria."
    icon = "icons/DistancesTransformation.svg"
    keywords = "distance transformation"

    class Inputs:
        distances = Input("Distances", DistMatrix)

    class Outputs:
        distances = Output("Distances", DistMatrix, dynamic=False)

    want_main_area = False
    resizing_enabled = False

    normalization_method = settings.Setting(0)
    inversion_method = settings.Setting(0)
    autocommit = settings.Setting(True)

    normalization_options = (
        ("No normalization", lambda x: x),
        ("To interval [0, 1]", lambda x: scale(x, min=0, max=1)),
        ("To interval [-1, 1]", lambda x: scale(x, min=-1, max=1)),
        ("Sigmoid function: 1/(1+exp(-X))", lambda x: 1/(1+np.exp(-x))),
    )

    inversion_options = (
        ("No inversion", lambda x: x),
        ("-X", lambda x: -x),
        ("1 - X", lambda x: 1-x),
        ("max(X) - X", lambda x: np.max(x) - x),
        ("1/X", lambda x: 1/x),
    )

    def __init__(self):
        super().__init__()

        self.data = None

        gui.radioButtons(self.controlArea, self, "normalization_method",
                         box="Normalization",
                         btnLabels=[x[0] for x in self.normalization_options],
                         callback=self._invalidate)

        gui.radioButtons(self.controlArea, self, "inversion_method",
                         box="Inversion",
                         btnLabels=[x[0] for x in self.inversion_options],
                         callback=self._invalidate)

        gui.auto_apply(self.buttonsArea, self, "autocommit")

    @Inputs.distances
    def set_data(self, data):
        self.data = data
        self.commit.now()

    @gui.deferred
    def commit(self):
        distances = self.data
        if distances is not None:
            # normalize
            norm = self.normalization_options[self.normalization_method][1]
            distances = norm(distances)

            # invert
            inv = self.inversion_options[self.inversion_method][1]
            distances = inv(distances)
        self.Outputs.distances.send(distances)

    def send_report(self):
        norm, normopt = self.normalization_method, self.normalization_options
        inv, invopt = self.inversion_method, self.inversion_options
        parts = []
        if inv:
            parts.append('inversion ({})'.format(invopt[inv][0]))
        if norm:
            parts.append('normalization ({})'.format(normopt[norm][0]))
        self.report_items(
            'Model parameters',
            {'Transformation': ', '.join(parts).capitalize() or 'None'})

    def _invalidate(self):
        self.commit.deferred()


if __name__ == "__main__":  # pragma: no cover
    import Orange.distance
    data = Orange.data.Table("iris")
    dist = Orange.distance.Euclidean(data)
    WidgetPreview(OWDistanceTransformation).run(dist)
